# Polychromatic is licensed under the GPLv3.
# Copyright (C) 2017-2024 Luke Horwell <code@horwell.me>
"""
Contains the UI localization for Polychromatic. Powered by GNU's gettext.
"""
import os
import gettext


class Locales(object):
    """
    Supports localisation throughout the application by utilising gettext.
    The "_" object is used for processing strings.
    """
    def __init__(self, force_locale=""):
        self.force_locale = force_locale
        self.locale = force_locale
        self.translation = None
        self._ = None

    def init(self):
        """
        Initialises translations for the application.

        Returns:
            gettext.translation() bound to an i18n variable.
        """
        is_relative = os.path.exists(os.path.join(os.path.dirname(__file__), "..", "data", "img"))
        relative_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "locale/"))

        # For development or a standalone "opt" build
        if is_relative and self.force_locale:
            self.translation = gettext.translation("polychromatic", localedir=relative_path, fallback=True, languages=[self.locale])
        elif is_relative:
            self.translation = gettext.translation("polychromatic", localedir=relative_path, fallback=True)

        # For packaged/system-wide installs
        elif not is_relative and self.force_locale:
            self.translation = gettext.translation("polychromatic", fallback=True, languages=[self.locale])
        else:
            self.translation = gettext.translation("polychromatic", fallback=True)

        self._ = self.translation.gettext
        return self._

    def get_current_locale(self):
        """
        Returns a string describing the current locale. E.g. "de" or "en_US".
        """
        if self.translation and "language" in self.translation.info():
            return self.translation.info()["language"]

        # Fallback to default
        return "en_GB"



#!/usr/bin/python3
#
# Polychromatic is licensed under the GPLv3.
# Copyright (C) 2017-2024 Luke Horwell <code@horwell.me>
#
"""
This module abstracts data from the OpenRazer Python library (and daemon)
and parses this for Polychromatic to use.

Project URL: https://github.com/openrazer/openrazer
"""
import glob
import os

import openrazer.client as rclient # pylint: disable=import-error

from .. import common
from ._backend import Backend as Backend


class OpenRazerBackend(Backend):
    """
    Integration with the OpenRazer 3.x Python library.

    Thoughout the module:
    - 'rdevice' refers to an openrazer.client.devices.RazerDevice object.
    - 'rzone' refers to an openrazer.client.fx.RazerFX (main) or
                           openrazer.client.fx.SingleLed object (e.g. logo)
    """
    def __init__(self, *args):
        super().__init__(*args)
        self.backend_id = "openrazer"
        self.name = "OpenRazer"
        self.logo = "openrazer.svg"
        self.version = rclient.__version__
        self.project_url = "https://openrazer.github.io"
        self.bug_url = "https://github.com/openrazer/openrazer/issues"
        self.releases_url = "https://github.com/openrazer/openrazer/releases"
        self.license = "GPLv2"

        # Variables for OpenRazer
        self.devman = None
        self.persistence_supported = True
        self.persistence_fallback_path = os.path.join(self.get_backend_storage_path(), "persistence")

        # Client Settings
        self.ripple_refresh_rate = 0.05
        self.load_client_overrides()

    def _reload_device_manager(self):
        """
        Returns a new instance of the OpenRazer Device Manager client. This
        establishes a connection to the daemon via D-Bus. The devices list will
        be up-to-date.

        If the daemon "service" is not running, this will usually start it.
        """
        self.debug("Connecting to daemon...")
        self.devman = rclient.DeviceManager()
        self.devman.sync_effects = False

    def init(self):
        """
        Summons the OpenRazer DeviceManager() daemon.
        """
        # Persistence API was introduced in OpenRazer 3.0.0
        if int(self.version.split(".")[0]) < 3:
            self.persistence_supported = False

        if not os.path.exists(self.persistence_fallback_path):
            os.makedirs(self.persistence_fallback_path)

        try:
            self._reload_device_manager()
            return True
        except Exception as e:
            self.debug("Failed: Got an exception initialising device manager!")
            return self.get_exception_as_string(e)

    def load_client_overrides(self):
        """
        Load any user-defined client settings that Polychromatic should use
        interfacing with the daemon. These are stored as individual files inside
        ~/.config/polychromatic/backends/openrazer/
        """
        def _get_override(filename, data_type, default):
            path = os.path.join(self.get_backend_storage_path(), filename)
            if not os.path.exists(path):
                return default

            with open(path, "r") as f:
                data = str(f.readline()).strip()

            try:
                output = data_type(data)
                self.debug("Setting client setting: {0} to {1}".format(filename, output))
                return output
            except ValueError:
                return default

        self.ripple_refresh_rate = _get_override("ripple_refresh_rate", float, 0.05)

    def get_unsupported_devices(self):
        """
        See Backend.get_unsupported_devices() and Backend.UnknownDeviceItem()

        Returns a list of PIDs of Razer hardware that is physically plugged in,
        but inaccessible by the daemon. Usually indicating the installation is
        incomplete or the device is not supported by the driver.
        """
        all_usb_pids = self.helpers.get_usb_pids_by_vid("1532")
        reg_pids = []
        unreg_pids = []

        # Get VIDs and PIDs from daemon to exclude later.
        if self.devman:
            for rdevice in self.devman.devices:
                vidpid = self._get_device_vid_pid(rdevice)
                reg_pids.append(vidpid.get("pid"))

        # Determine Razer PIDs that are not listed in the daemon
        for pid in all_usb_pids:
            if pid in reg_pids:
                continue

            # Ignore Kitty headphones duplicate: 1532:0521 [Headset], 1532:0F19 [Chroma] (#328)
            if pid == "0521":
                continue

            # Ignore devices with no RGB support
            elif pid in ["0511", "9051"]: # Razer Seiren, Razer Phone 2
                continue

            device = Backend.UnknownDeviceItem()
            device.name = "{0}:{1}".format("1532", pid)
            device.form_factor = self.get_form_factor()
            device.vid = "1532"
            device.pid = pid
            unreg_pids.append(device)

        return unreg_pids

    def get_devices(self):
        """
        See Backend.get_devices() and Backend.DeviceItem()
        """
        devices = []
        try:
            self._reload_device_manager()
        except Exception as e:
            return []
        for rdevice in self.devman.devices:
            devices.append(self._get_device(rdevice))
        return devices

    def get_device_by_name(self, name):
        """
        See Backend.get_device_by_name()
        """
        try:
            for rdevice in self.devman.devices:
                if rdevice.name == name:
                    return self._get_device(rdevice)
        except Exception as e:
            return self.get_exception_as_string(e)

    def get_device_by_serial(self, serial):
        """
        See Backend.get_device_by_serial()
        """
        try:
            for rdevice in self.devman.devices:
                if not rdevice.has("serial"):
                    continue
                if rdevice.serial == serial:
                    return self._get_device(rdevice)
        except Exception as e:
            return self.get_exception_as_string(e)

    def _get_device(self, rdevice):
        """
        Returns a Backend.DeviceItem() from OpenRazer's device object.
        """
        # A valid serial number is essential
        serial = ""
        if rdevice.has("serial"):
            serial = str(rdevice.serial)
        if len(serial) <= 2:
            serial = "".join(c for c in rdevice.name if c.isalnum()).upper()
            self.debug("Got bad serial for {0}! Using dummy serial: {1}".format(rdevice.name, serial))

        # Device details
        class OpenRazerDeviceItem(Backend.DeviceItem):
            def refresh(self):
                for zone in self.zones:
                    zone._persistence.refresh()
                    for option in zone.options:
                        option.refresh()
                if self.dpi:
                    self.dpi.refresh()

        device = OpenRazerDeviceItem()
        device._rdevice = rdevice
        device.name = str(rdevice.name)
        device.form_factor = self._get_form_factor(rdevice)
        device.real_image = self._get_device_image(rdevice)
        device.serial = serial
        device.monochromatic = self._is_device_monochromatic(rdevice)

        _vid_pid = self._get_device_vid_pid(rdevice)
        device.vid = _vid_pid.get("vid")
        device.pid = _vid_pid.get("pid")

        if rdevice.has("firmware_version"):
            device.firmware_version = str(rdevice.firmware_version)

        if rdevice.has("keyboard_layout"):
            device.keyboard_layout = str(rdevice.keyboard_layout)

        if rdevice.has("dpi") and not rdevice.has("available_dpi"):
            device.dpi = self._get_dpi_object(rdevice)

        if rdevice.has("battery"):
            device.battery = self._get_battery_object(rdevice)

        if rdevice.has("lighting_led_matrix"):
            device.matrix = self._get_matrix_object(rdevice, device)

        # Initialize zones
        device.zones = self._get_zone_objects(rdevice)
        main_zone = device.zones[0]

        # Add brightness & effects (per zone)
        for zone in device.zones:
            zone._persistence = self._get_persistence(self._map_zone_id_to_rzone(rdevice, zone), zone, serial)

            brightness = self._get_brightness_option(rdevice, zone)
            if brightness:
                zone.options.append(brightness)

            effects = self._get_effect_options(rdevice, zone)
            if effects:
                zone.options += effects

        workarounds = self._get_workaround_options(rdevice)
        if workarounds:
            main_zone.options = workarounds

        # Add other "main" options
        if rdevice.has("available_dpi"):
            device.dpi = None
            main_zone.options.append(self._get_dpi_fixed_object(rdevice))

        if rdevice.has("poll_rate"):
            main_zone.options.append(self._get_poll_rate_option(rdevice))

        if rdevice.has("game_mode_led"):
            main_zone.options.append(self._get_game_mode_option(rdevice))

        if rdevice.has("keyswitch_optimization"):
            main_zone.options.append(self._get_keyswitch_option(rdevice))

        if rdevice.has("battery"):
            main_zone.options += self._get_battery_options(rdevice)

        if rdevice.has("macro_mode_led_effect") and rdevice.type == "keyboard":
            device.has_programmable_keys = True

        if rdevice.has("macro_logic") and rdevice.type == "keyboard":
            device.has_macro_keys = True

        if rdevice.has("scroll_mode") or rdevice.has("scroll_acceleration") or rdevice.has("scroll_smart_reel"):
            main_zone.options += self._get_scroll_options(rdevice)

        return device

    def _get_persistence(self, rzone, zone, serial):
        """
        Returns OpenRazerPersistence() or OpenRazerPersistenceFallback()
        depending on this running version of OpenRazer.
        """
        if self.persistence_supported:
            return OpenRazerPersistence(rzone)

        return OpenRazerPersistenceFallback(zone.zone_id, serial, self.persistence_fallback_path)

    def _get_battery_object(self, rdevice):
        """
        Returns a Backend.DeviceItem.Battery object.
        """
        class Battery(Backend.DeviceItem.Battery):
            def __init__(self, rdevice):
                super().__init__()
                self._rdevice = rdevice
                self.is_charging = False
                self.percentage = -1

            def refresh(self):
                self.is_charging = self._rdevice.is_charging
                self.percentage = int(self._rdevice.battery_level)

        battery = Battery(rdevice)
        battery.refresh()

        # Some batteries require AA batteries
        if rdevice.name.find("Atheris") > 0:
            battery.is_removable = True

        return battery

    def _get_dpi_object(self, rdevice):
        """
        Returns a Backend.DeviceItem.DPI object with X/Y axis support.
        If the device uses "available_dpi", use _get_fixed_dpi_object() instead.

        Supports hardware that can save DPI intervals onto the hardware, if
        the "dpi_stages" capability is present.
        """
        class DPI(Backend.DeviceItem.DPI):
            def __init__(self, rdevice):
                super().__init__()
                self._rdevice = rdevice
                self.min = 100
                self.max = int(rdevice.max_dpi)

            def refresh(self):
                self.x = self._rdevice.dpi[0]
                self.y = self._rdevice.dpi[1]

            def set(self, x, y):
                self._rdevice.dpi = (int(x), int(y))

        class SyncDPI(DPI):
            def __init__(self, rdevice):
                super().__init__(rdevice)
                self.can_sync = True

            def sync(self, stages):
                """OpenRazer's "dpi_stages" setter expects: [active_stage, [stages: (x,y), (x,y)]"""
                stages = [(stage[0], stage[1]) for stage in stages]
                self._rdevice.dpi_stages = (1, stages)

        if rdevice.has("dpi_stages"):
            dpi = SyncDPI(rdevice)
        else:
            dpi = DPI(rdevice)

        # Determine DPI stages, or generate them if not known
        default_stages = {
            30000: [400, 800, 1600, 3200, 6400],
            16000: [800, 1800, 4500, 9000, 16000],
            8500: [800, 1600, 3200, 6400, 8500],
            8200: [800, 1800, 3200, 6400, 8200],
        }

        def _autogen_stage(value):
            """Guess a 'valid' DPI based on maximum as calculations may cause invalid range."""
            return round(value / 100) * 100

        try:
            stages = default_stages[dpi.max]
        except KeyError:
            stages = [
                _autogen_stage(dpi.max / 10),
                _autogen_stage(dpi.max / 8),
                _autogen_stage(dpi.max / 4),
                _autogen_stage(dpi.max / 2),
                int(dpi.max)
            ]

        # Polychromatic internally stores as a sublist of [X,Y] values
        dpi.default_stages = []
        for value in stages:
            dpi.default_stages.append([value, value])

        return dpi

    def _get_dpi_fixed_object(self, rdevice):
        """
        Returns a Backend.MultipleChoiceOption object as an alternate for DPI.
        This is used for devices that have a fixed DPI and do not support the
        'variable' slider.
        """
        current_dpi = int(rdevice.dpi[0])
        parameters = []

        for index, dpi in enumerate(list(rdevice.available_dpi)):
            param = Backend.Option.Parameter()
            param.data = int(dpi)
            param.label = str(dpi)
            param.active = True if dpi == current_dpi else False
            param.default = True if index == 0 else False
            parameters.append(param)

        class FixedDPIOption(Backend.MultipleChoiceOption):
            def __init__(self, rdevice, parameters):
                super().__init__()
                self._rdevice = rdevice
                self.uid = "fixed_dpi"
                self.parameters = parameters

            def refresh(self):
                current_dpi = int(self._rdevice.dpi[0])
                for param in self.parameters:
                    # Round up internally just in case DPI is not an exact value
                    param.active = True if round(param.data, -1) == round(current_dpi, -1) else False

            def apply(self, new_value):
                # Device only supports fixed DPI X values, such as DeathAdder 3.5G (#209)
                self._rdevice.dpi = (int(new_value), 0)

        fixed_dpi = FixedDPIOption(rdevice, parameters)
        fixed_dpi.label = self._("DPI")
        fixed_dpi.icon = self.get_icon("general", "dpi")

        return fixed_dpi

    def _get_matrix_object(self, rdevice, device):
        """
        Returns a Backend.DeviceItem.Matrix object.
        """
        class OpenRazerMatrix(Backend.DeviceItem.Matrix):
            def __init__(self, rdevice):
                self._rdevice = rdevice
                self.name = device.name
                self.form_factor_id = device.form_factor["id"]
                self.rows = int(rdevice.fx.advanced.rows)
                self.cols = int(rdevice.fx.advanced.cols)

            def set(self, x, y, red, green, blue):
                self._rdevice.fx.advanced.matrix[y, x] = (red, green, blue)

            def draw(self):
                self._rdevice.fx.advanced.draw()

            def clear(self):
                self._rdevice.fx.advanced.matrix.reset()

            def brightness(self):
                print("todo:stub:_get_matrix_object.brightness")

        class DeathStalkerMatrix(OpenRazerMatrix):
            """
            Alternate matrix implementation for Razer DeathStalker Chroma, which
            has a matrix of 12x1, but every second LED [2,4,6,8,10,12]
            physically blends with its previous LED [1,3,5,7,9,11], which messes
            up the lighting colours (#335)

            This matrix is virtual. It'll stretch LEDs by two for each one. Example:
                Virtual     Physical
                0       ->  0, 1
                5       ->  10, 11
            """
            def __init__(self, rdevice):
                super().__init__(rdevice)
                self.cols = 6

            def set(self, x, y, red, green, blue):
                self._rdevice.fx.advanced.matrix[y, (x * 2)] = (red, green, blue)
                self._rdevice.fx.advanced.matrix[y, (x * 2) + 1] = (red, green, blue)

        # OpenRazer changed this matrix after 3.1 (6 => 12)
        if rdevice.name == "Razer DeathStalker Chroma" and rdevice.fx.advanced.cols == 12:
            return DeathStalkerMatrix(rdevice)

        return OpenRazerMatrix(rdevice)

    def _get_zone_objects(self, rdevice):
        """
        Returns a list of Backend.DeviceItem.Zone objects.
        """
        zones = []
        device_name = str(rdevice.name)

        def _add_zone(zone_id, label):
            zone = Backend.DeviceItem.Zone()
            zone.zone_id = zone_id
            zone.label = label
            zone.icon = self.get_icon("zones", zone_id)
            zones.append(zone)

        # All devices have a 'main' base zone
        form_factor = self._get_form_factor(rdevice)
        zone = Backend.DeviceItem.Zone()
        zone.zone_id = "main"
        zone.label = form_factor["label"]
        zone.icon = form_factor["icon"]
        zones.append(zone)

        if rdevice.has("lighting_scroll") or rdevice.has("lighting_scroll_active"):
            _add_zone("scroll", self._("Scroll Wheel"))

        if rdevice.has("lighting_logo") or rdevice.has("lighting_logo_active"):
            # This zone may be more personalized for some devices
            zone = Backend.DeviceItem.Zone()
            zone.zone_id = "logo"
            zone.label = self._("Logo")
            zone.icon = self.get_icon("zones", "logo")

            if device_name.startswith("Razer Nex"):
                zone.label = self._("Hex Ring")
                zone.icon = self.get_icon("zones", "naga-hex-ring")

            if device_name.startswith("Razer Blade"):
                zone.label = self._("Laptop Lid")
                zone.icon = self.get_icon("zones", "blade-logo")

            zones.append(zone)

        if rdevice.has("lighting_left"):
            _add_zone("left", self._("Left"))
        if rdevice.has("lighting_right"):
            _add_zone("right", self._("Right"))
        if rdevice.has("lighting_backlight"):
            _add_zone("backlight", self._("Backlight"))
        if rdevice.has("lighting_charging"):
            _add_zone("charging", self._("Charging"))
        if rdevice.has("lighting_fast_charging"):
            _add_zone("fast_charging", self._("Fast Charging"))
        if rdevice.has("lighting_fully_charged"):
            _add_zone("fully_charged", self._("Fully Charging"))

        return zones

    def _get_form_factor(self, rdevice):
        """
        Convert the device type returned by OpenRazer to match one used within Polychromatic.
        """
        device_name = rdevice.name
        device_type = rdevice.type

        # Some of these 'device types' originate from legacy OpenRazer versions
        openrazer_to_poly = {
            "firefly": "mousemat",
            "tartarus": "keypad",
            "core": "gpu",
            "mug": "accessory"
        }

        if device_type in openrazer_to_poly:
            form_factor_id = openrazer_to_poly[device_type]
        else:
            form_factor_id = device_type

        if device_name.find("Base Station") != -1:
            form_factor_id = "stand"
        elif device_name.find("Blade") != -1:
            form_factor_id = "laptop"
        elif device_name.find("Core") != -1:
            form_factor_id = "gpu"
        elif device_name.find("Nommo") != -1:
            form_factor_id = "speaker"
        elif device_name.find("Raptor") != -1:
            form_factor_id = "display"

        return self.get_form_factor(form_factor_id)

    def _get_device_image(self, rdevice):
        """
        OpenRazer doesn't store device images, they are referenced by a URL.
        """
        try:
            # OpenRazer >= 2.9.0 (openrazer/openrazer#1127)
            return rdevice.device_image
        except AttributeError:
            # Backwards compatiblity with OpenRazer <= 2.8.0
            return rdevice.razer_urls["top_img"]
        except KeyError:
            return ""
        return ""

    def _get_device_vid_pid(self, rdevice):
        """
        Extracts VID:PID from the daemon's device object in list format: [VID,PID]

        In the event OpenRazer's _vid and _pid is inaccessible, then 0000 is returned.

        Returns:
            {vid, pid}      Success: A dictionary consisting of the VID and PID.
        """
        try:
            vid = str(hex(rdevice._vid))[2:].upper().rjust(4, '0')
            pid = str(hex(rdevice._pid))[2:].upper().rjust(4, '0')
        except Exception:
            self.debug("VID PID unavailable for " + rdevice.name + ". Using dummy ID.")
            vid = "0000"
            pid = "0000"

        return {
            "vid": vid,
            "pid": pid
        }

    def _is_device_monochromatic(self, device):
        """
        Returns a boolean to state whether the device supports per-lighting but
        only works with the 'green' value from RGB.
        """
        # Razer BlackWidow Ultimate keyboards only output "green" RGB
        if str(device.name).find("Ultimate") != -1 and device.type == "keyboard":
            return True

        return False

    def _map_zone_id_to_rzone(self, rdevice, zone):
        """
        Returns an object that directly references the OpenRazer's device "zone".
        """
        zone_to_device = {
            "main": rdevice.fx,
            "logo": rdevice.fx.misc.logo,
            "scroll": rdevice.fx.misc.scroll_wheel,
            "backlight": rdevice.fx.misc.backlight
        }

        # Introduced in OpenRazer 2.6.0, but not all devices list these.
        try:
            zone_to_device["left"] = rdevice.fx.misc.left
            zone_to_device["right"] = rdevice.fx.misc.right
        except KeyError:
            pass

        # Introduced in OpenRazer 3.0.0, but not all devices list these.
        try:
            zone_to_device["charging"] = rdevice.fx.misc.charging
            zone_to_device["fully_charged"] = rdevice.fx.misc.fully_charged
            zone_to_device["fast_charging"] = rdevice.fx.misc.fast_charging
        except KeyError:
            pass

        return zone_to_device[zone.zone_id]

    def _has_zone_capability(self, rdevice, zone, capability):
        """
        Returns a boolean whether the capability is available for the specified zone.
        For example, "active" for zone "logo" will check "lighting_logo_active"
        """
        zone_to_capability = {
            "main": "lighting",
            "logo": "lighting_logo",
            "scroll": "lighting_scroll",
            "backlight": "lighting_backlight",
            "left": "lighting_left",
            "right": "lighting_right",
            "charging": "lighting_charging",
            "fast_charging": "lighting_fast_charging",
            "fully_charged": "lighting_fully_charged",
        }

        # Brightness for the "root" (main) zone does not use the "lighting_" prefix.
        if capability == "brightness" and zone.zone_id == "main" and rdevice.has("brightness"):
            return True

        return rdevice.has(zone_to_capability[zone.zone_id] + "_" + capability)

    def _get_brightness_option(self, rdevice, zone):
        """
        Returns a Backend.Option derivative object based on the type of
        brightness for the specified zone and device.

        OpenRazer has two kinds of lighting:
            .brightness = a variable between 0 and 100.
            .active = an on/off state.

        Returns None if brightness is unsupported for the zone.
        """
        if rdevice.has("brightness") and zone.zone_id == "main":
            # This is provided in the root element, not .fx
            rzone = rdevice
        else:
            rzone = self._map_zone_id_to_rzone(rdevice, zone)

        # Device is a 'brightness' % variable
        if self._has_zone_capability(rdevice, zone, "brightness"):
            class BrightnessSlider(Backend.SliderOption):
                def __init__(self, rzone):
                    super().__init__()
                    self._rzone = rzone
                    self.uid = "brightness"
                    self.min = 0
                    self.max = 100
                    self.step = 5
                    self.suffix = "%"
                    self.suffix_plural = "%"

                def refresh(self):
                    self.value = int(round(self._rzone.brightness))

                def apply(self, new_value):
                    self._rzone.brightness = float(new_value)

            slider = BrightnessSlider(rzone)
            slider.label = self._("Brightness")
            slider.icon = self.get_icon("options", "brightness")
            return slider

        # Device uses an on/off state
        if self._has_zone_capability(rdevice, zone, "active"):
            class BrightnessToggle(Backend.ToggleOption):
                def __init__(self, rzone):
                    super().__init__()
                    self._rzone = rzone
                    self.uid = "brightness"

                def refresh(self):
                    self.active = True if self._rzone.active else False

                def apply(self, enabled):
                    self._rzone.active = enabled

            toggle = BrightnessToggle(rzone)
            toggle.label = self._("Brightness")
            toggle.icon = self.get_icon("options", "brightness")
            toggle.icon_enable = self.get_icon("params", "100")
            toggle.icon_disable = self.get_icon("params", "0")
            toggle.label_enable = self._("On")
            toggle.label_disable = self._("Off")
            toggle.label_toggle = self._("Enabled")
            return toggle

        # Device does not support this option
        return None

    def _get_effect_options(self, rdevice, zone):
        """
        Returns list of Backend.EffectOption objects by determining
        which options/parameters are available for this device and zone.
        """
        rzone = self._map_zone_id_to_rzone(rdevice, zone)
        options = []

        has_ripple = self._has_zone_capability(rdevice, zone, "ripple")
        has_ripple_random = self._has_zone_capability(rdevice, zone, "ripple_random")

        # There isn't a single 'lighting_breath' in the capabilities list
        has_breath_random = self._has_zone_capability(rdevice, zone, "breath_random")
        has_breath_mono = self._has_zone_capability(rdevice, zone, "breath_mono")
        has_breath_single = self._has_zone_capability(rdevice, zone, "breath_single")
        has_breath_dual = self._has_zone_capability(rdevice, zone, "breath_dual")
        has_breath_triple = self._has_zone_capability(rdevice, zone, "breath_triple")

        # There isn't a single 'lighting_starlight' in the capabilities list
        has_starlight_random = self._has_zone_capability(rdevice, zone, "starlight_random")
        has_starlight_single = self._has_zone_capability(rdevice, zone, "starlight_single")
        has_starlight_dual = self._has_zone_capability(rdevice, zone, "starlight_dual")

        if self._has_zone_capability(rdevice, zone, "none"):
            class NoneOption(Backend.EffectOption):
                def __init__(self, rzone, persistence):
                    super().__init__()
                    self._rzone = rzone
                    self._persistence = persistence
                    self.uid = "none"

                def refresh(self):
                    self.active = True if self._persistence.state["effect"] == "none" else False

                def apply(self, param=None):
                    self._rzone.none()
                    self._persistence.save("effect", "none")

            option = NoneOption(rzone, zone._persistence)
            option.label = self._("Off")
            option.icon = self.get_icon("params", "0")
            options.append(option)

        if self._has_zone_capability(rdevice, zone, "on"):
            class OnOption(Backend.EffectOption):
                def __init__(self, rzone, persistence):
                    super().__init__()
                    self._rzone = rzone
                    self._persistence = persistence
                    self.uid = "on"

                def refresh(self):
                    self.active = True if self._persistence.state["effect"] == "on" else False

                def apply(self, param=None):
                    self._rzone.on()
                    self._persistence.save("effect", "on")

            option = OnOption(rzone, zone._persistence)
            option.label = self._("On")
            option.icon = self.get_icon("params", "100")
            options.append(option)

        if self._has_zone_capability(rdevice, zone, "spectrum"):
            class SpectrumOption(Backend.EffectOption):
                def __init__(self, rzone, persistence):
                    super().__init__()
                    self._rzone = rzone
                    self._persistence = persistence
                    self.uid = "spectrum"

                def refresh(self):
                    self.active = True if self._persistence.state["effect"] == "spectrum" else False

                def apply(self, param=None):
                    self._rzone.spectrum()
                    self._persistence.save("effect", "spectrum")

            option = SpectrumOption(rzone, zone._persistence)
            option.label = self._("Spectrum")
            option.icon = self.get_icon("options", "spectrum")
            options.append(option)

        if self._has_zone_capability(rdevice, zone, "wave"):
            class WaveOption(Backend.EffectOption):
                def __init__(self, rzone, persistence):
                    super().__init__()
                    self._rzone = rzone
                    self._persistence = persistence
                    self.uid = "wave"

                def refresh(self):
                    self.active = True if self._persistence.state["effect"] == "wave" else False
                    for param in self.parameters:
                        param.active = True if self._persistence.state["wave_dir"] == param.data else False

                def apply(self, direction):
                    # direction: 1 or 2
                    self._rzone.wave(int(direction))
                    self._persistence.save("effect", "wave")
                    self._persistence.save("wave_dir", str(direction))

            option = WaveOption(rzone, zone._persistence)
            option.label = self._("Wave")
            option.icon = self.get_icon("options", "wave")

            direction_1 = Backend.Option.Parameter()
            direction_1.data = 1

            direction_2 = Backend.Option.Parameter()
            direction_2.data = 2
            direction_2.default = True

            # Change parameter labels depending on orientation/device
            if rdevice.type == "mouse":
                direction_1.label = self._("Up")
                direction_1.icon  = self.get_icon("params", "up")
                direction_2.label = self._("Down")
                direction_2.icon  = self.get_icon("params", "down")

            elif rdevice.type == "mousemat":
                direction_1.label = self._("Clockwise")
                direction_1.icon  = self.get_icon("params", "clock")
                direction_2.label = self._("Anti-clockwise")
                direction_2.icon  = self.get_icon("params", "anticlock")

            else:
                direction_1.label = self._("Right")
                direction_1.icon = self.get_icon("params", "right")
                direction_2.label = self._("Left")
                direction_2.icon = self.get_icon("params", "left")

            option.parameters = [direction_2, direction_1]
            options.append(option)

        if self._has_zone_capability(rdevice, zone, "wheel"):
            class WheelOption(Backend.EffectOption):
                def __init__(self, rzone, persistence):
                    super().__init__()
                    self._rzone = rzone
                    self._persistence = persistence
                    self.uid = "wheel"

                def refresh(self):
                    self.active = True if self._persistence.state["effect"] == "wheel" else False
                    for param in self.parameters:
                        param.active = True if self._persistence.state["wave_dir"] == param.data else False

                def apply(self, direction):
                    # direction: 1 or 2
                    self._rzone.wheel(int(direction))
                    self._persistence.save("effect", "wheel")
                    self._persistence.save("wave_dir", str(direction))

            option = WheelOption(rzone, zone._persistence)
            option.label = self._("Wheel")
            option.icon = self.get_icon("options", "wheel")

            direction_1 = Backend.Option.Parameter()
            direction_1.data = 1
            direction_1.label = self._("Clockwise")
            direction_1.icon = self.get_icon("params", "right")

            direction_2 = Backend.Option.Parameter()
            direction_2.data = 2
            direction_2.default = True
            direction_2.label = self._("Anti-clockwise")
            direction_2.icon = self.get_icon("params", "left")

            option.parameters = [direction_2, direction_1]
            options.append(option)

        if has_ripple or has_ripple_random:
            class RippleOption(Backend.EffectOption):
                def __init__(self, rzone, persistence):
                    super().__init__()
                    self._rzone = rzone
                    self._persistence = persistence
                    self.uid = "ripple"
                    self.colours = persistence.state["colours"]

                def refresh(self):
                    current_effect = self._persistence.state["effect"]
                    self.active = True if current_effect in ["ripple", "rippleRandomColour"] else False
                    for param in self.parameters:
                        if param.data == "random":
                            param.active = True if current_effect == "rippleRandomColour" else False
                        elif param.data == "single":
                            param.active = True if current_effect == "ripple" else False
                    self.colours = self._persistence.state["colours"]

                def apply(self, ripple_type):
                    if str(ripple_type) == "random":
                        self._rzone.ripple_random()
                        self._persistence.save("effect", "rippleRandomColour")
                    elif str(ripple_type) == "single":
                        rgb = common.hex_to_rgb(self.colours[0])
                        self._rzone.ripple(rgb[0], rgb[1], rgb[2])
                        self._persistence.save("effect", "ripple")
                        self._persistence.save("colour_1", self.colours[0])

            option = RippleOption(rzone, zone._persistence)
            option.label = self._("Ripple")
            option.icon = self.get_icon("options", "ripple")

            if has_ripple_random:
                random = Backend.Option.Parameter()
                random.data = "random"
                random.label = self._("Random")
                random.icon = self.get_icon("params", "random")
                option.parameters.append(random)

            if has_ripple:
                single = Backend.Option.Parameter()
                single.data = "single"
                single.label = self._("Single")
                single.icon = self.get_icon("params", "single")
                single.colours_required = 1
                single.default = True
                option.parameters.append(single)

            options.append(option)

        if self._has_zone_capability(rdevice, zone, "reactive"):
            class ReactiveOption(Backend.EffectOption):
                def __init__(self, rzone, persistence):
                    super().__init__()
                    self._rzone = rzone
                    self._persistence = persistence
                    self.uid = "reactive"
                    self.colours_required = 1
                    self.colours = self._persistence.state["colours"]

                def refresh(self):
                    self.active = True if self._persistence.state["effect"] == "reactive" else False
                    for param in self.parameters:
                        param.active = True if self._persistence.state["speed"] == param.data else False
                    self.colours = self._persistence.state["colours"]

                def apply(self, speed):
                    rgb = common.hex_to_rgb(self.colours[0])
                    self._rzone.reactive(rgb[0], rgb[1], rgb[2], int(speed))
                    self._persistence.save("effect", "reactive")
                    self._persistence.save("speed", str(speed))
                    self._persistence.save("colour_1", self.colours[0])

            option = ReactiveOption(rzone, zone._persistence)
            option.label = self._("Reactive")
            option.icon = self.get_icon("options", "reactive")

            fast = Backend.Option.Parameter()
            fast.data = 1
            fast.label = self._("Fast (0.5s)")
            fast.icon = self.get_icon("params", "fast")
            fast.colours_required = 1

            medium = Backend.Option.Parameter()
            medium.data = 2
            medium.label = self._("Medium (1s)")
            medium.default = True
            medium.colours_required = 1

            slow = Backend.Option.Parameter()
            slow.data = 3
            slow.label = self._("Slow (1.5s)")
            slow.colours_required = 1

            vslow = Backend.Option.Parameter()
            vslow.data = 4
            vslow.label = self._("Very Slow (2s)")
            vslow.icon = self.get_icon("params", "slow")
            vslow.colours_required = 1

            for param in [fast, medium, slow, vslow]:
                option.parameters.append(param)

            options.append(option)

        if self._has_zone_capability(rdevice, zone, "blinking"):
            # Chroma Mug Holder is the only one to have it in the "main" zone, but there's no Python API call.
            # API only exposes for 'logo' and 'scroll'. Some mice use it.
            if zone.zone_id in ["logo", "scroll"]:
                class BlinkingOption(Backend.EffectOption):
                    def __init__(self, rzone, persistence):
                        super().__init__()
                        self._rzone = rzone
                        self._persistence = persistence
                        self.uid = "blinking"
                        self.colours_required = 1
                        self.colours = self._persistence.state["colours"]

                    def refresh(self):
                        self.active = True if self._persistence.state["effect"] == "blinking" else False
                        self.colours = self._persistence.state["colours"]

                    def apply(self, param=None):
                        rgb = common.hex_to_rgb(self.colours[0])
                        self._rzone.blinking(rgb[0], rgb[1], rgb[2])
                        self._persistence.save("effect", "blinking")
                        self._persistence.save("colour_1", self.colours[0])

                option = BlinkingOption(rzone, zone._persistence)
                option.label = self._("Blinking")
                option.icon = self.get_icon("options", "blinking")
                options.append(option)

        if self._has_zone_capability(rdevice, zone, "static"):
            class StaticOption(Backend.EffectOption):
                def __init__(self, rzone, persistence):
                    super().__init__()
                    self._rzone = rzone
                    self._persistence = persistence
                    self.uid = "static"
                    self.colours_required = 1
                    self.colours = self._persistence.state["colours"]

                def refresh(self):
                    self.active = True if self._persistence.state["effect"] == "static" else False
                    self.colours = self._persistence.state["colours"]

                def apply(self, param=None):
                    rgb = common.hex_to_rgb(self.colours[0])
                    self._rzone.static(rgb[0], rgb[1], rgb[2])
                    self._persistence.save("effect", "static")
                    self._persistence.save("colour_1", self.colours[0])

            option = StaticOption(rzone, zone._persistence)
            option.label = self._("Static")
            option.icon = self.get_icon("options", "static")
            options.append(option)

        if has_breath_random or has_breath_mono or has_breath_single or has_breath_dual or has_breath_triple:
            class BreathOption(Backend.EffectOption):
                def __init__(self, rzone, persistence):
                    super().__init__()
                    self._rzone = rzone
                    self._persistence = persistence
                    self.uid = "breath"
                    self.colours = self._persistence.state["colours"]

                def refresh(self):
                    current_effect = self._persistence.state["effect"]
                    if not current_effect.startswith("breath"):
                        self.active = False
                        return
                    self.active = True
                    current_breath_type = current_effect.split("breath")[1].lower()
                    for param in self.parameters:
                        param.active = True if current_breath_type == param.data else False
                    self.colours = self._persistence.state["colours"]

                def apply(self, breath_type):
                    rgb = []
                    for colour in self.colours:
                        rgb.append(common.hex_to_rgb(colour))

                    if breath_type == "random":
                        self._rzone.breath_random()
                        self._persistence.save("effect", "breathRandom")
                    elif breath_type == "mono":
                        self._rzone.breath_mono()
                        self._persistence.save("effect", "breathMono")
                    elif breath_type == "single":
                        self._rzone.breath_single(rgb[0][0], rgb[0][1], rgb[0][2])
                        self._persistence.save("effect", "breathSingle")
                        self._persistence.save("colour_1", self.colours[0])
                    elif breath_type == "dual":
                        self._rzone.breath_dual(rgb[0][0], rgb[0][1], rgb[0][2],
                                                rgb[1][0], rgb[1][1], rgb[1][2])
                        self._persistence.save("effect", "breathDual")
                        self._persistence.save("colour_1", self.colours[0])
                        self._persistence.save("colour_2", self.colours[1])
                    elif breath_type == "triple":
                        self._rzone.breath_triple(rgb[0][0], rgb[0][1], rgb[0][2],
                                                  rgb[1][0], rgb[1][1], rgb[1][2],
                                                  rgb[2][0], rgb[2][1], rgb[2][2])
                        self._persistence.save("effect", "breathTriple")
                        self._persistence.save("colour_1", self.colours[0])
                        self._persistence.save("colour_2", self.colours[1])
                        self._persistence.save("colour_3", self.colours[2])
                    else:
                        raise KeyError("Unknown breath type: " + breath_type)

            option = BreathOption(rzone, zone._persistence)
            option.label = self._("Breath")
            option.icon = self.get_icon("options", "breath")

            if has_breath_random:
                random = Backend.Option.Parameter()
                random.data = "random"
                random.label = self._("Random")
                random.icon = self.get_icon("params", "random")
                option.parameters.append(random)

            if has_breath_mono:
                mono = Backend.Option.Parameter()
                mono.data = "mono"
                mono.label = self._("Mono")
                mono.icon = self.get_icon("params", "mono")
                option.parameters.append(mono)

            if has_breath_single:
                single = Backend.Option.Parameter()
                single.data = "single"
                single.label = self._("Single")
                single.icon = self.get_icon("params", "single")
                single.colours_required = 1
                single.default = True
                option.parameters.append(single)

            if has_breath_dual:
                dual = Backend.Option.Parameter()
                dual.data = "dual"
                dual.label = self._("Dual")
                dual.icon = self.get_icon("params", "dual")
                dual.colours_required = 2
                option.parameters.append(dual)

            if has_breath_triple:
                triple = Backend.Option.Parameter()
                triple.data = "triple"
                triple.label = self._("Triple")
                triple.icon = self.get_icon("params", "triple")
                triple.colours_required = 3
                option.parameters.append(triple)

            options.append(option)

        if has_starlight_random or has_starlight_single or has_starlight_dual:
            class StarlightOption(Backend.EffectOption):
                def __init__(self, rzone, persistence):
                    super().__init__()
                    self._rzone = rzone
                    self._persistence = persistence
                    self.uid = "starlight"
                    self.colours = self._persistence.state["colours"]

                def refresh(self):
                    current_effect = self._persistence.state["effect"]
                    if not current_effect.startswith("starlight"):
                        self.active = False
                        return
                    self.active = True
                    current_starlight = current_effect.split("starlight")[1].lower()
                    current_speed = self._persistence.state["speed"]
                    self.active = True if current_effect.startswith("starlight") else False
                    for param in self.parameters:
                        param.active = False
                        starlight_type, starlight_speed = param.data.split(":")
                        if current_starlight == starlight_type and str(current_speed) == starlight_speed:
                            param.active = True
                    self.colours = self._persistence.state["colours"]

                def apply(self, data):
                    # Param Example: "random:2" for a Medium (2) Random Starlight
                    starlight_type = data.split(":")[0]
                    starlight_speed = int(data.split(":")[1])

                    rgb = []
                    for colour in self.colours:
                        rgb.append(common.hex_to_rgb(colour))

                    if starlight_type == "random":
                        self._rzone.starlight_random(starlight_speed)
                        self._persistence.save("effect", "starlightRandom")
                    elif starlight_type == "single":
                        self._rzone.starlight_single(rgb[0][0], rgb[0][1], rgb[0][2], starlight_speed)
                        self._persistence.save("colour_1", self.colours[0])
                        self._persistence.save("effect", "starlightSingle")
                    elif starlight_type == "dual":
                        self._rzone.starlight_dual(rgb[0][0], rgb[0][1], rgb[0][2],
                                                   rgb[1][0], rgb[1][1], rgb[1][2], starlight_speed)
                        self._persistence.save("colour_1", self.colours[0])
                        self._persistence.save("colour_2", self.colours[1])
                        self._persistence.save("effect", "starlightDual")
                    else:
                        raise KeyError("Unknown starlight parameter:" + str(data))
                    self._persistence.save("speed", starlight_speed)

            option = StarlightOption(rzone, zone._persistence)
            option.label = self._("Starlight")
            option.icon = self.get_icon("options", "starlight")

            speeds = {
                1: self._("Fast"),
                2: self._("Medium"),
                3: self._("Slow"),
            }

            if has_starlight_random:
                for speed in speeds.keys():
                    random = Backend.Option.Parameter()
                    random.data = "random:" + str(speed)
                    random._speed = speed
                    random.label = "{0} ({1})".format(self._("Random"), speeds[speed])
                    random.icon = self.get_icon("params", "random")
                    option.parameters.append(random)

            if has_starlight_single:
                for speed in speeds.keys():
                    single = Backend.Option.Parameter()
                    single.data = "single:" + str(speed)
                    single._speed = speed
                    single.label = "{0} ({1})".format(self._("Single"), speeds[speed])
                    single.icon = self.get_icon("params", "single")
                    single.colours_required = 1
                    single.default = True
                    option.parameters.append(single)

            if has_starlight_dual:
                for speed in speeds.keys():
                    dual = Backend.Option.Parameter()
                    dual.data = "dual:" + str(speed)
                    dual._speed = speed
                    dual.label = "{0} ({1})".format(self._("Dual"), speeds[speed])
                    dual.icon = self.get_icon("params", "dual")
                    dual.colours_required = 2
                    option.parameters.append(dual)

            options.append(option)

        return options

    def _get_workaround_options(self, rdevice):
        """
        If applicable, return a list of option objects that workaround the
        OpenRazer Python library due to bugs in the API.

        #1: Devices speaking the "BW2013" protocol can't set pulsate or static.
            - The latter doesn't appear as a capability either. These devices do not
            accept parameters or colours.
            - As a workaround, bypass the pylib and echo directly to the sysfs driver.
            - See also: #345, openrazer/openrazer#1575
        """
        try:
            if "razer.device.lighting.bw2013" in rdevice._available_features.keys():
                vidpid = self._get_device_vid_pid(rdevice)
                persistence = OpenRazerPersistenceFallback("main", rdevice.serial, self.persistence_fallback_path)

                try:
                    matrix_file_pulsate = glob.glob("/sys/bus/hid/drivers/razer*/*{0}:{1}*/matrix_effect_pulsate".format(vidpid["vid"], vidpid["pid"]), recursive=True)[0]
                    matrix_file_static = glob.glob("/sys/bus/hid/drivers/razer*/*{0}:{1}*/matrix_effect_static".format(vidpid["vid"], vidpid["pid"]), recursive=True)[0]
                except IndexError:
                    # Check the fake driver instead
                    matrix_file_pulsate = glob.glob("/tmp/**/*{0}:{1}*/matrix_effect_pulsate".format(vidpid["vid"], vidpid["pid"]), recursive=True)[0]
                    matrix_file_static = glob.glob("/tmp/**/*{0}:{1}*/matrix_effect_static".format(vidpid["vid"], vidpid["pid"]), recursive=True)[0]

                class PulsateOptionBW2013(Backend.EffectOption):
                    def __init__(self, persistence, sysfs_path):
                        super().__init__()
                        self._persistence = persistence
                        self.sysfs_path = sysfs_path
                        self.uid = "pulsate"

                    def refresh(self):
                        self.active = True if self._persistence.state["effect"] == "pulsate" else False

                    def apply(self, param=None):
                        with open(self.sysfs_path, "w") as f:
                            f.write("1")
                        self._persistence.save("effect", "pulsate")

                class StaticOptionBW2013(Backend.EffectOption):
                    def __init__(self, persistence, sysfs_path):
                        super().__init__()
                        self.sysfs_path = sysfs_path
                        self._persistence = persistence
                        self.uid = "static"

                    def refresh(self):
                        self.active = True if self._persistence.state["effect"] == "static" else False

                    def apply(self, param=None):
                        with open(self.sysfs_path, "w") as f:
                            f.write("1")
                        self._persistence.save("effect", "static")

                pulsate = PulsateOptionBW2013(persistence, matrix_file_pulsate)
                pulsate.label = self._("Pulsate")
                pulsate.icon = self.get_icon("options", "pulsate")

                static = StaticOptionBW2013(persistence, matrix_file_static)
                static.label = self._("Static")
                static.icon = self.get_icon("options", "static")

                self.debug("Using sysfs workaround for Pulsate/Static")
                return [pulsate, static]

        except AttributeError:
            self.debug("Can't check _available_features as not supported by library. Omitting any workarounds.")

        return None

    def _get_poll_rate_option(self, rdevice):
        """
        Returns a Backend.Option derivative object for setting a mouse's poll rate.
        """
        current_rate = int(rdevice.poll_rate)
        parameters = []

        # OpenRazer <= 3.1.0 were hardcoded (not exposed via API)
        supported_poll_rates = [125, 500, 1000]

        # OpenRazer >= 3.2.0 provides the list
        if rdevice.has("supported_poll_rates"):
            supported_poll_rates = rdevice.supported_poll_rates

        for rate in supported_poll_rates:
            param = Backend.Option.Parameter()
            param.data = rate
            param.active = True if current_rate == rate else False

            # 500 Hz  = 2 millisecond latency
            # 1000 Hz = 1 millisecond latency
            # 2000 Hz = 0.5 millisecond latency
            msecs = float(1000 / rate)
            param.label = self._("X Hz (Y msec latency)").replace("X", str(rate)).replace("Y", str(int(msecs) if msecs.is_integer() else msecs))

            if rate > 1000:
                param.icon = self.get_icon("params", "poll_hyper")
            elif rate > 500:
                param.icon = self.get_icon("params", "poll_high")
            elif rate < 500:
                param.icon = self.get_icon("params", "poll_low")
            else:
                param.icon = self.get_icon("params", "poll_mid")

            parameters.append(param)

        class PollRateOption(Backend.MultipleChoiceOption):
            def __init__(self, rdevice, parameters):
                super().__init__()
                self._rdevice = rdevice
                self.uid = "poll_rate"
                self.parameters = parameters

            def refresh(self):
                current_rate = int(self._rdevice.poll_rate)
                for param in self.parameters:
                    param.active = True if param.data == current_rate else False

            def apply(self, new_value):
                self._rdevice.poll_rate = int(new_value)

        poll_rate = PollRateOption(rdevice, parameters)
        poll_rate.label = self._("Poll Rate")
        poll_rate.icon = self.get_icon("options", "poll_rate")
        return poll_rate

    def _get_game_mode_option(self, rdevice):
        """
        Returns a Backend.Option derivative object for the hardware's game mode feature.
        """
        class GameModeOption(Backend.ToggleOption):
            def __init__(self, rdevice):
                super().__init__()
                self._rdevice = rdevice
                self.uid = "game_mode"

            def refresh(self):
                self.active = True if rdevice.game_mode_led else False

            def apply(self, enabled):
                self._rdevice.game_mode_led = enabled

        option = GameModeOption(rdevice)
        option.label =  self._("Game Mode")
        option.label_toggle = self._("Disable Alt+Tab, Alt+F4 and Win keys")
        option.icon = self.get_icon("options", "game_mode")
        option.icon_enable = self.get_icon("options", "game_mode")
        option.icon_disable = self.get_icon("options", "game_mode_off")
        return option

    def _get_battery_options(self, rdevice):
        """
        Returns a list of Backend.Option derivative objects for power saving features.

        In OpenRazer >= 3.2.0, low power and sleep mode are exposed as individual capabilities.
        """
        options = []
        persistence = OpenRazerPersistenceFallback("main", rdevice.serial, self.persistence_fallback_path)

        # This is the amount of time before the device enters "sleep mode"
        if rdevice.has("get_idle_time") or rdevice.has("set_idle_time"):
            class IdleTimeOptionSetOnly(Backend.SliderOption):
                def __init__(self, rdevice, persistence):
                    # Device stores idle time in seconds. Present as minutes.
                    super().__init__()
                    self._rdevice = rdevice
                    self._persistence = persistence
                    self.uid = "idle_time"
                    self.min = 1
                    self.max = 15

                def refresh(self):
                    self.value = int(int(self._persistence.get("idle_time")) / 60)

                def apply(self, new_value):
                    self._rdevice.set_idle_time(int(new_value) * 60)
                    self._persistence.save("idle_time", int(new_value) * 60)

            class IdleTimeOptionSetGet(IdleTimeOptionSetOnly):
                def refresh(self):
                    self.value = int(self._rdevice.get_idle_time() / 60)

            if rdevice.has("idle_time") or rdevice.has("get_idle_time"):
                idle_time = IdleTimeOptionSetGet(rdevice, persistence)
            else:
                idle_time = IdleTimeOptionSetOnly(rdevice, persistence)

            idle_time.label = self._("Sleep mode after")
            idle_time.icon = self.get_icon("options", "sleep")
            idle_time.suffix = ' ' + self._("minute")
            idle_time.suffix_plural = ' ' + self._("minutes")
            options.append(idle_time)

        # This is the battery percentage before the device enters a low power mode.
        if rdevice.has("get_low_battery_threshold") or rdevice.has("set_low_battery_threshold"):
            class LowBatteryThresholdOptionSetOnly(Backend.SliderOption):
                def __init__(self, rdevice, persistence):
                    super().__init__()
                    self._rdevice = rdevice
                    self._persistence = persistence
                    self.uid = "low_battery_threshold"
                    self.min = 1
                    self.max = 100
                    self.suffix = "%"
                    self.suffix_plural = "%"

                def refresh(self):
                    self.value = int(self._persistence.get("low_battery_threshold"))

                def apply(self, new_value):
                    self._rdevice.set_low_battery_threshold(int(new_value))
                    self._persistence.save("low_battery_threshold", int(new_value))

            class LowBatteryThresholdOptionSetGet(LowBatteryThresholdOptionSetOnly):
                def refresh(self):
                    self.value = int(self._rdevice.get_low_battery_threshold())

            if rdevice.has("low_battery_threshold") or rdevice.has("get_low_battery_threshold"):
                low_power = LowBatteryThresholdOptionSetGet(rdevice, persistence)
            else:
                low_power = LowBatteryThresholdOptionSetOnly(rdevice, persistence)

            low_power.label = self._("Enter low power at")
            low_power.icon = self.get_icon("options", "low_battery")
            options.append(low_power)

        return options

    def _get_scroll_options(self, rdevice):
        """
        Additional mouse features supported by Razer Basilisk V3 since OpenRazer 3.3.0.
        """
        options = []

        if rdevice.has("scroll_mode"):
            tactile = Backend.Option.Parameter()
            tactile.data = 0
            tactile.label = self._("Tactile")
            tactile.default = True

            free_spin = Backend.Option.Parameter()
            free_spin.data = 1
            free_spin.label = self._("Free Spin")

            class ScrollMode(Backend.MultipleChoiceOption):
                def __init__(self, rdevice):
                    super().__init__()
                    self._rdevice = rdevice

                def refresh(self):
                    tactile.active = rdevice.scroll_mode == 0
                    free_spin.active = rdevice.scroll_mode == 1

                def apply(self, value):
                    self._rdevice.scroll_mode = int(value)

            scroll_mode = ScrollMode(rdevice)
            scroll_mode.uid = "scroll_mode"
            scroll_mode.label = self._("Scroll Mode")
            scroll_mode.icon = self.get_icon("devices", "mouse")
            scroll_mode.parameters = [tactile, free_spin]
            scroll_mode.refresh()
            options.append(scroll_mode)

        if rdevice.has("scroll_acceleration"):
            class ScrollAcceleration(Backend.ToggleOption):
                def __init__(self, rdevice):
                    super().__init__()
                    self._rdevice = rdevice

                def refresh(self):
                    self.active = self._rdevice.scroll_acceleration == True

                def apply(self, state):
                    self._rdevice.scroll_acceleration = state

            scroll_accel = ScrollAcceleration(rdevice)
            scroll_accel.uid = "scroll_accel"
            scroll_accel.label = self._("Scroll Acceleration")
            scroll_accel.icon = self.get_icon("devices", "mouse")
            scroll_accel.label_toggle = self._("Enable scroll acceleration")
            scroll_accel.refresh()
            options.append(scroll_accel)

        if rdevice.has("scroll_smart_reel"):
            class SmartReel(Backend.ToggleOption):
                def __init__(self, rdevice):
                    super().__init__()
                    self._rdevice = rdevice

                def refresh(self):
                    self.active = self._rdevice.scroll_smart_reel == True

                def apply(self, state):
                    self._rdevice.scroll_smart_reel = state

            smart_reel = SmartReel(rdevice)
            smart_reel.uid = "scroll_smart_reel"
            smart_reel.label = self._("Smart Reel")
            smart_reel.icon = self.get_icon("devices", "mouse")
            smart_reel.label_toggle = self._("Enable smart reel")
            smart_reel.refresh()
            options.append(smart_reel)

        return options

    def _get_keyswitch_option(self, rdevice):
        """
        Additional mouse features supported by Razer Huntsman V2 since OpenRazer 3.4.0.
        """
        typing = Backend.Option.Parameter()
        typing.data = False
        typing.label = self._("Typing (Increased debounce delay)")
        typing.default = True

        gaming = Backend.Option.Parameter()
        gaming.data = True
        gaming.label = self._("Gaming (Faster triggering)")

        class KeyswitchOptimisation(Backend.MultipleChoiceOption):
            def __init__(self, rdevice):
                super().__init__()
                self._rdevice = rdevice

            def refresh(self):
                typing.active = rdevice.keyswitch_optimization == False
                gaming.active = rdevice.keyswitch_optimization == True

            def apply(self, value):
                self._rdevice.keyswitch_optimization = value == True

        option = KeyswitchOptimisation(rdevice)
        option.uid = "keyswitch_optimisation"
        option.label = self._("Optimise for")
        option.icon = self.get_icon("devices", "keyboard")
        option.parameters = [typing, gaming]
        option.refresh()
        return option

    def restart(self):
        """
        Immediately restart the daemon process.
        """
        import time

        # Stop any process running
        self.debug("Running: openrazer-daemon -s")
        os.system("openrazer-daemon -s")

        # Give chance to stop, but kill to be sure.
        self.debug("Waiting for openrazer-daemon to stop (2s)...")
        time.sleep(2)
        os.system("killall openrazer-daemon")

        # Start again
        self.debug("Running: openrazer-daemon")
        os.system("openrazer-daemon")

        self.debug("Waiting for openrazer-daemon to start (2s)...")
        time.sleep(2)


class OpenRazerPersistence(object):
    """
    Use OpenRazer's persistence API introduced in v3.0.0. Each 'fx' zone contains
    variables that the daemon uses for storing the last effect, parameters and colours.

    Due to a daemon bug, there is no way to tell if a device does not support/need
    persistence, so we must fail gracefully (#294, openrazer/openrazer#1380)
    """
    state = {
        "effect": "spectrum",
        "colours": ["#00FF00", "#FF0000", "#0000FF"],
        "wave_dir": 1,
        "speed": 2,
    }

    def __init__(self, rzone):
        self.rzone = rzone

    def _convert_colour_bytes(self, rzone):
        """
        Convert the daemon's '.colors' hex output to a list consisting of #RRGGBB strings.
        """
        input_hex = rzone.colors.hex()

        if len(input_hex) >= 6:
            primary_hex = input_hex[:6]

        if len(input_hex) >= 12:
            secondary_hex = input_hex[6:12]

        if len(input_hex) >= 18:
            tertiary_hex = input_hex[12:18]

        return [
            "#" + primary_hex,
            "#" + secondary_hex,
            "#" + tertiary_hex
        ]

    def refresh(self):
        try:
            self.state["effect"] = str(self.rzone.effect)
            self.state["wave_dir"] = int(self.rzone.wave_dir)
            self.state["speed"] = int(self.rzone.speed)
            self.state["colours"] = self._convert_colour_bytes(self.rzone)
        except Exception:
            pass

    def save(self, key, value):
        # Not applicable. The daemon is taking care of storing persistence.
        return


class OpenRazerPersistenceFallback(OpenRazerPersistence):
    """
    Use a file-based persistence for backwards compatibility (<= 2.9.0)
    """
    def __init__(self, zone_id, serial, path):
        self.zone_id = zone_id
        self.serial = serial
        self.persistence_path = path

        # "colours" will be saved as separate files
        self.state["colour_1"] = self.state["colours"][0]
        self.state["colour_2"] = self.state["colours"][1]
        self.state["colour_3"] = self.state["colours"][2]

    def _get_key_path(self, key):
        return os.path.join(self.persistence_path, f"{self.serial}_{self.zone_id}_{key}")

    def _get_data(self, key, data_type):
        file_path = self._get_key_path(key)
        if os.path.exists(file_path):
            with open(file_path) as f:
                self.state[key] = data_type(f.readline())

    def refresh(self):
        self._get_data("effect", str)
        self._get_data("wave_dir", int)
        self._get_data("speed", int)
        self._get_data("colour_1", str)
        self._get_data("colour_2", str)
        self._get_data("colour_3", str)
        self.state["colours"] = [
            self.state["colour_1"],
            self.state["colour_2"],
            self.state["colour_3"],
        ]

    def get(self, key):
        file_path = self._get_key_path(key)
        if os.path.exists(file_path):
            with open(file_path) as f:
                return str(f.readline())
        return "0"

    def save(self, key, value):
        if not os.path.exists(self.persistence_path):
            os.makedirs(self.persistence_path)
        with open(self._get_key_path(key), "w") as f:
            f.write(str(value))



# Polychromatic is licensed under the GPLv3.
# Copyright (C) 2020-2024 Luke Horwell <code@horwell.me>
"""
This module controls the visual editor shared between layered and scripted effects.
"""

from .. import common
from .. import effects
from ..fx import FX
from .. import locales
from .. import preferences as pref
from .. import procpid
from . import shared
from . import effects as controller_effects

from ..qt.flowlayout import FlowLayout as QFlowLayout

import copy
import json
import os
import time

from PyQt5 import QtCore
from PyQt5.QtCore import Qt, QRect, QItemSelectionModel, QThread, QSize, QUrl
from PyQt5.QtGui import QIcon, QPixmap, QFont, QColor
from PyQt5.QtWidgets import QWidget, QPushButton, QToolButton, QMessageBox, \
                            QListWidget, QTreeWidget, QLabel, QComboBox, \
                            QTreeWidgetItem, QMenu, QDialog, QDialogButtonBox, \
                            QButtonGroup, QLineEdit, QTextEdit, QCheckBox, \
                            QGroupBox, QRadioButton, QMainWindow, QAction, \
                            QDockWidget, QMenuBar, QToolBar, QStatusBar, \
                            QTableWidget, QWidget, QVBoxLayout, QHBoxLayout, \
                            QScrollArea, QSpinBox, QDesktopWidget, QColorDialog
from PyQt5.QtWebEngineWidgets import QWebEngineView

# Visual mode within the WebView editor
VISUAL_MODE_ADD = 1
VISUAL_MODE_DEL = 2
VISUAL_MODE_PICK = 3


class VisualEffectEditor(shared.TabData):
    """
    Editor for modifying layered or animated effects. The GUI is shared between
    a "Layered" and "Sequence" type of effect.
    """
    def __init__(self, appdata, fileman, save_path):
        """
        Load the effect from file for editing and show the editor window. The
        file is expected to have been initialised (via EffectMetadataEditor) first.

        Params:
            appdata         ApplicationData() object
            fileman         fileman.FlatFileManagement() based object
            effect_path     Path to save data
        """
        super().__init__(appdata)

        self.fileman = fileman
        self.save_path = save_path
        self.data = fileman.get_item(save_path)
        self.modified = False
        self.alive = True
        self.current_colour = "#00FF00"
        self.current_tool = VISUAL_MODE_ADD
        self.layer_labels = self._get_layer_labels()
        self.webview_zoom_level = 1.0
        self.fx = FX()

        # Effect type
        self.effect_type = self.data["type"]
        self.layered_effect = self.effect_type == effects.TYPE_LAYERED
        self.sequence_effect = self.effect_type == effects.TYPE_SEQUENCE

        # Fail if the file moved, or was deleted
        if type(self.data) != dict:
            self._show_file_error()
            return

        # Session variables
        self.suppress_delete_prompt = appdata.preferences["editor"]["suppress_confirm_dialog"]

        # Layered Effects Only
        self.current_layer = 0

        # Sequence Effects Only
        self.current_frame = 0
        self.playback_paused = True
        self.playback_thread = None

        # Load window
        self.window = shared.get_ui_widget(appdata, "editor", QMainWindow)
        self.window.closeEvent = self.closeEvent
        self.menubar = self.window.findChild(QMenuBar, "MenuBar")
        self.statusbar = self.window.findChild(QStatusBar, "StatusBar")
        self.toolbar = self.window.findChild(QToolBar, "MainToolbar")

        # Visual Editor ("webview")
        self.webview = self.window.findChild(QWebEngineView, "MatrixWebView")
        self.action_info_graphic = self.window.findChild(QAction, "MapGraphic")
        self.action_info_grid_x = self.window.findChild(QAction, "MapDimensionX")
        self.action_info_grid_y = self.window.findChild(QAction, "MapDimensionY")

        # -- DeviceRender() is assigned to this variable in init_controls()
        self.device_renderer = None

        # Device (live preview)
        self.live_preview = self.preferences["editor"]["live_preview"]
        self.device = None
        self.matrix = None
        self.init_device_preview()

        # Set up webview paths
        self.webview.page().profile().setCachePath(pref.path.webview_cache)
        self.webview.page().profile().setPersistentStoragePath(pref.path.webview_cache)

        # Assign widgets to variables
        # -- Menu Bar/Toolbar - File
        self.action_save_apply = self.window.findChild(QAction, "actionSaveApply")
        self.action_save = self.window.findChild(QAction, "actionSave")
        self.action_revert = self.window.findChild(QAction, "actionRevert")
        self.action_exit = self.window.findChild(QAction, "actionExit")

        # -- Menu Bar/Toolbar - Edit
        self.menu_edit = self.window.findChild(QMenu, "menuEdit")

        # -- -- Layer only
        self.action_new_layer = self.window.findChild(QAction, "actionNew_Layer")
        self.action_delete_layer = self.window.findChild(QAction, "actionDelete_Layer")
        self.action_duplicate_layer = self.window.findChild(QAction, "actionDuplicate_Layer")
        self.action_layer_up = self.window.findChild(QAction, "actionMove_Layer_Up")
        self.action_layer_down = self.window.findChild(QAction, "actionMove_Layer_Down")

        # -- -- Frame only
        self.action_new_frame = self.window.findChild(QAction, "actionNew_Frame")
        self.action_delete_frame = self.window.findChild(QAction, "actionDelete_Frame")
        self.action_clone_frame = self.window.findChild(QAction, "actionClone_Frame")
        self.action_frame_left = self.window.findChild(QAction, "actionMove_Frame_Left")
        self.action_frame_right = self.window.findChild(QAction, "actionMove_Frame_Right")

        self.menu_shift = self.window.findChild(QMenu, "menuShiftMenu")
        self.action_shift_left = self.window.findChild(QAction, "actionShift_Left")
        self.action_shift_right = self.window.findChild(QAction, "actionShift_Right")
        self.action_shift_up = self.window.findChild(QAction, "actionShift_Up")
        self.action_shift_down = self.window.findChild(QAction, "actionShift_Down")
        self.action_edit_metadata = self.window.findChild(QAction, "actionEditMetadata")
        self.action_edit_triggers = self.window.findChild(QAction, "actionEditTriggers")

        # -- Menu Bar/Toolbar - View
        self.action_view_device_graphic = self.window.findChild(QAction, "actionShow_as_Device_Graphic")
        self.action_view_grid = self.window.findChild(QAction, "actionShow_as_Grid")
        self.action_view_key_labels = self.window.findChild(QAction, "actionShowKeyLabels")
        self.action_zoom_in = self.window.findChild(QAction, "actionZoomIn")
        self.action_zoom_out = self.window.findChild(QAction, "actionZoomOut")
        self.action_zoom_reset = self.window.findChild(QAction, "actionZoomReset")

        # -- Menu Bar/Toolbar - Tools
        self.menu_tools = self.window.findChild(QMenu, "menuTools")
        self.tool_draw = self.window.findChild(QAction, "actionDraw")
        self.tool_eraser = self.window.findChild(QAction, "actionEraser")
        self.tool_picker = self.window.findChild(QAction, "actionColourPicker")

        # -- Menu Bar/Dock - Colours
        self.menu_colours = self.window.findChild(QMenu, "menuColours")

        # -- Menu Bar/Dock - Playback
        self.playback_menu = self.window.findChild(QMenu, "menuPlayback")
        self.playback_jump_start = self.window.findChild(QAction, "actionJumpStart")
        self.playback_jump_end = self.window.findChild(QAction, "actionJumpEnd")
        self.playback_prev = self.window.findChild(QAction, "actionPreviousFrame")
        self.playback_next = self.window.findChild(QAction, "actionNextFrame")
        self.playback_play = self.window.findChild(QAction, "actionPlay")
        self.playback_stop = self.window.findChild(QAction, "actionStop")
        self.playback_loop = self.window.findChild(QAction, "actionLoop")

        # -- Menu Bar - Help
        self.action_help_online = self.window.findChild(QAction, "actionOnlineHelp")
        self.action_help_website = self.window.findChild(QAction, "actionWebsite")
        self.action_help_relnotes = self.window.findChild(QAction, "actionReleaseNotes")
        self.action_help_report_bug = self.window.findChild(QAction, "actionReportBug")
        self.action_help_donate = self.window.findChild(QAction, "actionDonate")
        self.action_help_about = self.window.findChild(QAction, "actionAbout")

        # -- Docks
        self.dock_layers = self.window.findChild(QDockWidget, "LayersDock")
        self.dock_properties = self.window.findChild(QDockWidget, "LayerPropertiesDock")
        self.dock_colours = self.window.findChild(QDockWidget, "ColoursDock")
        self.dock_frames = self.window.findChild(QDockWidget, "FramesDock")

        # -- Layer Dock
        self.layer_tree = self.window.findChild(QTreeWidget, "LayerTree")
        self.btn_layer_delete = self.window.findChild(QPushButton, "LayerDelete")
        self.btn_layer_duplicate = self.window.findChild(QPushButton, "LayerDuplicate")
        self.btn_layer_move_down = self.window.findChild(QPushButton, "LayerMoveDown")
        self.btn_layer_move_up = self.window.findChild(QPushButton, "LayerMoveUp")
        self.btn_layer_new = self.window.findChild(QPushButton, "LayerNew")

        # -- Frame Dock
        self.frame_table = self.window.findChild(QTableWidget, "FramesTable")

        self.btn_frame_new = self.window.findChild(QPushButton, "NewFrame")
        self.btn_frame_delete = self.window.findChild(QPushButton, "DeleteFrame")
        self.btn_frame_clone = self.window.findChild(QPushButton, "CloneFrame")
        self.btn_frame_move_left = self.window.findChild(QPushButton, "MoveFrameLeft")
        self.btn_frame_move_right = self.window.findChild(QPushButton, "MoveFrameRight")

        self.btn_playback_jump_start = self.window.findChild(QToolButton, "PlaybackJumpStart")
        self.btn_playback_jump_end = self.window.findChild(QToolButton, "PlaybackJumpEnd")
        self.btn_playback_play = self.window.findChild(QToolButton, "PlaybackStart")
        self.btn_playback_stop = self.window.findChild(QToolButton, "PlaybackStop")
        self.btn_playback_prev = self.window.findChild(QToolButton, "PlaybackPrev")
        self.btn_playback_next = self.window.findChild(QToolButton, "PlaybackNext")
        self.btn_playback_loop = self.window.findChild(QCheckBox, "PlaybackLoop")
        self.spinner_playback_fps = self.window.findChild(QSpinBox, "PlaybackFPS")
        self.status_frame = QLabel()

        # -- Colours Menu Bar / Dock
        self.saved_colours = self.window.findChild(QWidget, "SavedColoursWidget")

        self.current_colour_block = self.window.findChild(QWidget, "CurrentColourBlock")
        self.current_colour_textbox = self.window.findChild(QLineEdit, "CurrentColourTextBox")
        self.btn_change_colour = self.window.findChild(QPushButton, "ChangeCurrentColour")

        self.label_hue = self.window.findChild(QLabel, "HueLabel")
        self.label_saturation = self.window.findChild(QLabel, "SaturationLabel")
        self.label_lightness = self.window.findChild(QLabel, "LightnessLabel")
        self.btn_hue_increase = self.window.findChild(QPushButton, "HueIncrease")
        self.btn_hue_decrease = self.window.findChild(QPushButton, "HueDecrease")
        self.btn_saturation_increase = self.window.findChild(QPushButton, "SaturationIncrease")
        self.btn_saturation_decrease = self.window.findChild(QPushButton, "SaturationDecrease")
        self.btn_lightness_increase = self.window.findChild(QPushButton, "LightnessIncrease")
        self.btn_lightness_decrease = self.window.findChild(QPushButton, "LightnessDecrease")

        self.action_colours_picker = self.window.findChild(QAction, "actionOpenColourPicker")
        self.action_colours_hue_increase = self.window.findChild(QAction, "actionIncreaseHue")
        self.action_colours_hue_decrease = self.window.findChild(QAction, "actionDecreaseHue")
        self.action_colours_saturation_increase = self.window.findChild(QAction, "actionIncreaseSaturation")
        self.action_colours_saturation_decrease = self.window.findChild(QAction, "actionDecreaseSaturation")
        self.action_colours_lightness_increase = self.window.findChild(QAction, "actionIncreaseLightness")
        self.action_colours_lightness_decrease = self.window.findChild(QAction, "actionDecreaseLightness")
        self.action_colours_manage = self.window.findChild(QAction, "actionManageSavedColours")

        # Font should be applied to dock widgets
        self.docks = [self.dock_layers, self.dock_properties, self.dock_colours, self.dock_frames]
        if not self.appdata.system_qt_theme:
            for dock in self.docks:
                font = QFont("Play", 10, 0)
                dock.setFont(font)

        # Connect buttons/actions
        # -- File
        self.action_save_apply.triggered.connect(self.save_and_close)
        self.action_save.triggered.connect(self.save_now)
        self.action_revert.triggered.connect(self.reload_from_file)
        self.action_exit.triggered.connect(self.closeEvent)

        # -- Edit (Layered)
        self.action_new_layer.triggered.connect(self.new_layer)
        self.action_delete_layer.triggered.connect(self.delete_layer)
        self.action_duplicate_layer.triggered.connect(self.duplicate_layer)
        self.action_layer_up.triggered.connect(self.raise_layer)
        self.action_layer_down.triggered.connect(self.lower_layer)

        self.btn_layer_new.clicked.connect(self.new_layer)
        self.btn_layer_delete.clicked.connect(self.delete_layer)
        self.btn_layer_duplicate.clicked.connect(self.duplicate_layer)
        self.btn_layer_move_up.clicked.connect(self.raise_layer)
        self.btn_layer_move_down.clicked.connect(self.lower_layer)

        # -- Edit (Sequence)
        self.action_new_frame.triggered.connect(self.new_frame)
        self.action_delete_frame.triggered.connect(self.delete_frame)
        self.action_clone_frame.triggered.connect(self.clone_frame)
        self.action_frame_left.triggered.connect(self.shift_frame_left)
        self.action_frame_right.triggered.connect(self.shift_frame_right)

        # -- Edit (Shared)
        self.action_shift_left.triggered.connect(self.shift_all_left)
        self.action_shift_right.triggered.connect(self.shift_all_right)
        self.action_shift_up.triggered.connect(self.shift_all_up)
        self.action_shift_down.triggered.connect(self.shift_all_down)
        self.action_edit_metadata.triggered.connect(self.edit_metadata)
        self.action_edit_triggers.triggered.connect(self.edit_triggers)

        # -- View
        self.action_view_device_graphic.triggered.connect(self.view_device_graphic)
        self.action_view_grid.triggered.connect(self.view_device_grid)
        self.action_view_key_labels.triggered.connect(self.view_key_labels)
        self.action_zoom_in.triggered.connect(self.zoom_in)
        self.action_zoom_out.triggered.connect(self.zoom_out)
        self.action_zoom_reset.triggered.connect(self.zoom_reset)

        # -- Tools
        self.tool_draw.triggered.connect(self.select_mode_draw)
        self.tool_eraser.triggered.connect(self.select_mode_eraser)
        self.tool_picker.triggered.connect(self.select_mode_picker)

        # -- Colours
        self.action_colours_picker.triggered.connect(self.open_colour_picker)
        self.action_colours_hue_increase.triggered.connect(self.set_colour_increase_hue)
        self.action_colours_hue_decrease.triggered.connect(self.set_colour_decrease_hue)
        self.action_colours_saturation_increase.triggered.connect(self.set_colour_increase_saturation)
        self.action_colours_saturation_decrease.triggered.connect(self.set_colour_decrease_saturation)
        self.action_colours_lightness_increase.triggered.connect(self.set_colour_increase_lightness)
        self.action_colours_lightness_decrease.triggered.connect(self.set_colour_decrease_lightness)
        self.action_colours_manage.triggered.connect(self.edit_saved_colours)

        # -- Playback (Sequence only)
        if self.sequence_effect:
            self._init_playback_controls()

        # -- Help
        menubar = self.appdata.menubar
        self.action_help_online.triggered.connect(menubar.online_help)
        self.action_help_website.triggered.connect(menubar.polychromatic_website)
        self.action_help_relnotes.triggered.connect(menubar.polychromatic_release_notes)
        self.action_help_report_bug.triggered.connect(menubar.polychromatic_report_bug)
        self.action_help_donate.triggered.connect(menubar.polychromatic_donate)
        self.action_help_about.triggered.connect(menubar.about_polychromatic)

        # -- Dock: Frames
        self.frame_table.itemSelectionChanged.connect(self.open_frame)
        self.btn_frame_new.clicked.connect(self.new_frame)
        self.btn_frame_delete.clicked.connect(self.delete_frame)
        self.btn_frame_clone.clicked.connect(self.clone_frame)
        self.btn_frame_move_left.clicked.connect(self.shift_frame_left)
        self.btn_frame_move_right.clicked.connect(self.shift_frame_right)

        # -- Dock: Colours (click)
        self.btn_change_colour.clicked.connect(self.open_colour_picker)
        self.current_colour_textbox.textChanged.connect(self.set_colour_by_textbox)
        self.btn_hue_increase.clicked.connect(self.set_colour_increase_hue)
        self.btn_saturation_increase.clicked.connect(self.set_colour_increase_saturation)
        self.btn_lightness_increase.clicked.connect(self.set_colour_increase_lightness)
        self.btn_hue_decrease.clicked.connect(self.set_colour_decrease_hue)
        self.btn_saturation_decrease.clicked.connect(self.set_colour_decrease_saturation)
        self.btn_lightness_decrease.clicked.connect(self.set_colour_decrease_lightness)

        # -- Dock: Colours (scroll)
        for widget in [self.label_hue, self.btn_hue_increase, self.btn_hue_decrease]:
            def _hue_onscroll(evt):
                if evt.angleDelta().y() > 0:
                    self.set_colour_increase_hue()
                else:
                    self.set_colour_decrease_hue()
            widget.wheelEvent = _hue_onscroll

        for widget in [self.label_saturation, self.btn_saturation_increase, self.btn_saturation_decrease]:
            def _saturation_onscroll(evt):
                if evt.angleDelta().y() > 0:
                    self.set_colour_increase_saturation()
                else:
                    self.set_colour_decrease_saturation()
            widget.wheelEvent = _saturation_onscroll

        for widget in [self.label_lightness, self.btn_lightness_decrease, self.btn_lightness_increase]:
            def _lightness_onscroll(evt):
                if evt.angleDelta().y() > 0:
                    self.set_colour_increase_lightness()
                else:
                    self.set_colour_decrease_lightness()
            widget.wheelEvent = _lightness_onscroll

        # Override Behaviours
        # -- Scroll on frames (sequence only)
        def frame_table_wheelEvent(event):
            direction = 1 if int(event.angleDelta().y()) > 0 else -1
            self.frame_table.selectColumn(self.frame_table.currentColumn() + direction)
        self.frame_table.wheelEvent = frame_table_wheelEvent

        # Menu Bar Icons
        if not appdata.system_qt_theme:
            # -- File
            self.action_save_apply.setIcon(self.widgets.get_icon_qt("general", "ok"))
            self.action_save.setIcon(self.widgets.get_icon_qt("general", "save"))
            self.action_revert.setIcon(self.widgets.get_icon_qt("general", "reset"))
            self.action_exit.setIcon(self.widgets.get_icon_qt("general", "exit"))

            # -- Edit
            self.action_new_layer.setIcon(self.widgets.get_icon_qt("general", "new"))
            self.action_delete_layer.setIcon(self.widgets.get_icon_qt("general", "delete"))
            self.action_duplicate_layer.setIcon(self.widgets.get_icon_qt("general", "clone"))
            self.action_layer_up.setIcon(self.widgets.get_icon_qt("general", "move-up"))
            self.action_layer_down.setIcon(self.widgets.get_icon_qt("general", "move-down"))

            self.action_new_frame.setIcon(self.widgets.get_icon_qt("general", "new"))
            self.action_delete_frame.setIcon(self.widgets.get_icon_qt("general", "delete"))
            self.action_clone_frame.setIcon(self.widgets.get_icon_qt("general", "clone"))
            self.action_frame_left.setIcon(self.widgets.get_icon_qt("general", "move-left"))
            self.action_frame_right.setIcon(self.widgets.get_icon_qt("general", "move-right"))

            self.menu_shift.setIcon(self.widgets.get_icon_qt("general", "move-up"))
            self.action_shift_left.setIcon(self.widgets.get_icon_qt("general", "move-left"))
            self.action_shift_right.setIcon(self.widgets.get_icon_qt("general", "move-right"))
            self.action_shift_up.setIcon(self.widgets.get_icon_qt("general", "move-up"))
            self.action_shift_down.setIcon(self.widgets.get_icon_qt("general", "move-down"))
            self.action_edit_metadata.setIcon(self.widgets.get_icon_qt("general", "properties"))
            self.action_edit_triggers.setIcon(self.widgets.get_icon_qt("general", "triggers"))

            # -- View
            self.action_view_device_graphic.setIcon(self.widgets.get_icon_qt("general", "devices"))
            self.action_view_grid.setIcon(self.widgets.get_icon_qt("general", "matrix"))
            self.action_view_key_labels.setIcon(self.widgets.get_icon_qt("devices", "keyboard"))
            self.action_zoom_in.setIcon(self.widgets.get_icon_qt("effects", "zoom-in"))
            self.action_zoom_out.setIcon(self.widgets.get_icon_qt("effects", "zoom-out"))
            self.action_zoom_reset.setIcon(self.widgets.get_icon_qt("effects", "zoom-reset"))

            # -- Tools
            self.tool_draw.setIcon(self.widgets.get_icon_qt("effects", "paint"))
            self.tool_eraser.setIcon(self.widgets.get_icon_qt("effects", "eraser"))
            self.tool_picker.setIcon(self.widgets.get_icon_qt("effects", "picker"))

            # -- Colours
            self.action_colours_picker.setIcon(self.widgets.get_icon_qt("general", "palette"))
            self.action_colours_hue_increase.setIcon(self.widgets.get_icon_qt("effects", "more"))
            self.action_colours_saturation_increase.setIcon(self.widgets.get_icon_qt("effects", "more"))
            self.action_colours_lightness_increase.setIcon(self.widgets.get_icon_qt("effects", "more"))
            self.action_colours_hue_decrease.setIcon(self.widgets.get_icon_qt("effects", "minus"))
            self.action_colours_saturation_decrease.setIcon(self.widgets.get_icon_qt("effects", "minus"))
            self.action_colours_lightness_decrease.setIcon(self.widgets.get_icon_qt("effects", "minus"))
            self.action_colours_manage.setIcon(self.widgets.get_icon_qt("general", "edit"))

            # -- Playback
            self.playback_jump_start.setIcon(self.widgets.get_icon_qt("effects", "rewind"))
            self.playback_jump_end.setIcon(self.widgets.get_icon_qt("effects", "fast-forward"))
            self.playback_prev.setIcon(self.widgets.get_icon_qt("effects", "step-backward"))
            self.playback_next.setIcon(self.widgets.get_icon_qt("effects", "step-forward"))
            self.playback_play.setIcon(self.widgets.get_icon_qt("effects", "play"))
            self.playback_stop.setIcon(self.widgets.get_icon_qt("effects", "stop"))
            self.playback_loop.setIcon(self.widgets.get_icon_qt("effects", "repeat"))

            # -- Help
            self.action_help_online.setIcon(self.widgets.get_icon_qt("general", "external"))

        # Dock Button Icons
        if not appdata.system_qt_theme:
            # -- Layer Dock
            self.btn_layer_delete.setIcon(self.widgets.get_icon_qt("general", "delete"))
            self.btn_layer_duplicate.setIcon(self.widgets.get_icon_qt("general", "clone"))
            self.btn_layer_move_down.setIcon(self.widgets.get_icon_qt("general", "move-down"))
            self.btn_layer_move_up.setIcon(self.widgets.get_icon_qt("general", "move-up"))
            self.btn_layer_new.setIcon(self.widgets.get_icon_qt("general", "new"))

            # -- Frame Dock
            self.btn_frame_new.setIcon(self.widgets.get_icon_qt("general", "new"))
            self.btn_frame_delete.setIcon(self.widgets.get_icon_qt("general", "delete"))
            self.btn_frame_clone.setIcon(self.widgets.get_icon_qt("general", "clone"))
            self.btn_frame_move_left.setIcon(self.widgets.get_icon_qt("general", "move-left"))
            self.btn_frame_move_right.setIcon(self.widgets.get_icon_qt("general", "move-right"))

            self.btn_playback_jump_start.setIcon(self.widgets.get_icon_qt("effects", "rewind"))
            self.btn_playback_jump_end.setIcon(self.widgets.get_icon_qt("effects", "fast-forward"))
            self.btn_playback_play.setIcon(self.widgets.get_icon_qt("effects", "play"))
            self.btn_playback_stop.setIcon(self.widgets.get_icon_qt("effects", "stop"))
            self.btn_playback_prev.setIcon(self.widgets.get_icon_qt("effects", "step-backward"))
            self.btn_playback_next.setIcon(self.widgets.get_icon_qt("effects", "step-forward"))

            # -- Colour Dock
            self.btn_change_colour.setIcon(self.widgets.get_icon_qt("general", "palette"))
            self.btn_hue_increase.setIcon(self.widgets.get_icon_qt("effects", "more"))
            self.btn_saturation_increase.setIcon(self.widgets.get_icon_qt("effects", "more"))
            self.btn_lightness_increase.setIcon(self.widgets.get_icon_qt("effects", "more"))
            self.btn_hue_decrease.setIcon(self.widgets.get_icon_qt("effects", "minus"))
            self.btn_saturation_decrease.setIcon(self.widgets.get_icon_qt("effects", "minus"))
            self.btn_lightness_decrease.setIcon(self.widgets.get_icon_qt("effects", "minus"))

        # Showtime!
        self._init_window()
        self.init_controls()

        # FIXME: Hide incomplete features
        for widget in [self.window.findChild(QAction, "actionEditTriggers")]:
            widget.setDisabled(True)
            widget.setVisible(False)

    def _init_window(self):
        """
        Prepare the window for the type of effect that'll be edited.

        Layered Effects use these docks/interface elements:
            - Layers
            - Properties

        Sequence Effects use:
            - Colours
            - Frames
        """
        if self.layered_effect:
            # Remove frame-specific options
            self.tool_picker.deleteLater()
            self.action_new_frame.deleteLater()
            self.action_delete_frame.deleteLater()
            self.action_clone_frame.deleteLater()
            self.action_frame_left.deleteLater()
            self.action_frame_right.deleteLater()
            self.playback_menu.deleteLater()

        elif self.sequence_effect:
            # Remove layer-specific options
            self.action_new_layer.deleteLater()
            self.action_delete_layer.deleteLater()
            self.action_duplicate_layer.deleteLater()
            self.action_layer_up.deleteLater()
            self.action_layer_down.deleteLater()
            self.dock_layers.deleteLater()
            self.dock_properties.deleteLater()

        # Window properties & geometry
        self.appdata.main_window._set_initial_window_position(self.window, "editor")

        if self.window.windowIcon().isNull():
            self.window.setWindowIcon(QIcon(common.get_icon("general", "effects")))

        # -- If opening in the center, make better use of the screen space
        centered = self.preferences["controller"]["window_behaviour"] == pref.WINDOW_BEHAVIOUR_CENTER
        already_small = self.window.height() > 1024 and self.window.width() > 768
        if centered and not already_small:

            # DEPRECATED: QDesktopWidget(), but some distros ship an older Qt version (5.8, 5.12)
            qt_ver = QtCore.PYQT_VERSION_STR.split(".")
            if qt_ver[0] == "5" and int(qt_ver[1]) <= 14:
                screen = QDesktopWidget().availableGeometry()
            else:
                # For >= Qt 5.15
                screen = self.window.screen().availableGeometry()

            frame = self.window.frameGeometry()
            prefer_height = screen.height() - 175
            prefer_width = screen.width() - 275

            # Position in the center, as usual
            center = screen.center()
            frame.setHeight(prefer_height)
            frame.setWidth(prefer_width)
            frame.moveCenter(center)
            self.window.resize(prefer_width, prefer_height)
            self.window.move(frame.topLeft())

        # Toolbar style
        self.widgets.set_toolbar_style(self.toolbar)

        for button in [self.btn_frame_new, self.btn_frame_delete, self.btn_frame_clone,
                       self.btn_frame_move_left, self.btn_frame_move_right]:
            self.widgets.set_toolbar_style(button)

        # Adjust docks for optimum space
        if self.layered_effect:
            self.dock_layers.adjustSize()
            self.dock_properties.adjustSize()
        elif self.sequence_effect:
            self.dock_colours.adjustSize()
            self.dock_frames.adjustSize()

    def init_controls(self):
        """
        Populate the controls for the effect being loaded.
        """
        # Reset session variables
        self.set_modified(False)
        self.current_layer = 0
        self.current_frame = 0

        # Show Key Labels
        hide_key_labels = self.preferences["editor"]["hide_key_labels"]
        if not hide_key_labels:
            self.action_view_key_labels.setChecked(True)

        if not self.data["map_device_icon"] in ["keyboard", "keypad"]:
            self.action_view_key_labels.setEnabled(False)

        # Prepare visual editor
        rows = self.data["map_rows"]
        cols = self.data["map_cols"]
        graphic_filename = self.data["map_graphic"]
        self.device_renderer = DeviceRenderer(self.appdata, self, self.webview, self.init_editor, graphic_filename, rows, cols, hide_key_labels)
        self.select_mode_draw()
        graphic_name = self.device_renderer.graphic_name

        # Device Graphic / Grid
        self.action_view_device_graphic.setEnabled(True if self.data["map_graphic"] else False)
        self.action_info_graphic.setText(graphic_name)
        self.action_info_grid_x.setText(self._("Columns: []").replace("[]", str(rows)))
        self.action_info_grid_y.setText(self._("Rows: []").replace("[]", str(cols)))

        if not self.data["map_graphic"]:
            self.action_info_graphic.setText(self._("(Not Set)"))

    def init_editor(self):
        """
        Populate the visual editor (grid/graphic) when the webpage is ready.
        """
        self.dbg.stdout("Reading data...", self.dbg.action, 1)

        if self.layered_effect:
            # Populate layer dock
            self.load_layers()

            # Open the first layer and populate properties dock
            self.open_layer()

        elif self.sequence_effect:
            # Populate colours dock
            self.load_colours()

            # Populate frame dock
            self.load_frames()
            self.open_frame()

            # Show the current/total frame position in status bar
            self.statusbar.addPermanentWidget(self.status_frame)

        self.dbg.stdout("Load complete.", self.dbg.success, 1)
        self.window.show()

    def init_device_preview(self):
        """
        Initialise the objects for previewing the device on physical hardware.
        """
        if not self.live_preview:
            return

        # Prepare device object
        device_name = self.data["map_device"]
        self.dbg.stdout("Preparing live preview on '{0}'...".format(device_name), self.dbg.action, 1)
        self.device = self.middleman.get_device_by_name(device_name)

        # Inform the user if the backends are not ready
        if not self.appdata.ready:
            self.dbg.stdout("Backends not ready! Preview not possible yet.", self.dbg.error, 1)
            retry = False

            def do_retry():
                nonlocal retry
                retry = True

            def do_ignore():
                self.dbg.stdout("Loading editor without live preview.", self.dbg.action, 1)

            self.widgets.open_dialog(self.widgets.dialog_warning,
                                        self._("Backend Error"),
                                        self._("Live preview isn't possible yet as the backends haven't finished initialising."),
                                        self._("They might be ready now, try again. Or, choose Ignore to continue without a live preview on physical hardware."),
                                        buttons=[QMessageBox.Retry, QMessageBox.Ignore],
                                        actions={
                                            QMessageBox.Retry: do_retry,
                                            QMessageBox.Ignore: do_ignore
                                        })

            if retry:
                return self.init_device_preview()

        # Inform the user if attempting to load non-existent device
        if not self.device:
            self.dbg.stdout("Device not found! No preview.", self.dbg.error, 1)
            self.widgets.open_dialog(self.widgets.dialog_warning,
                                        self._("Missing Device"),
                                        self._("This effect is designed for a device named '[]', but it wasn't detected.").replace("[]", device_name),
                                        self._("Live preview has been disabled for this session."))
            return

        if self.device and self.device.matrix:
            self.matrix = self.device.matrix

        # Inform the user if attempting to use an unsupported device
        if self.device and not self.matrix:
            self.dbg.stdout("Device not supported! No preview.", self.dbg.error, 1)
            self.widgets.open_dialog(self.widgets.dialog_error,
                                     self._("File Error"),
                                     self._("The device '[]' is assigned to this effect, but does not support individual addressable LEDs.").replace("[]", device_name),
                                     self._("Live preview has been disabled for this session."))
            self.device = None
            return

        # Stop any other effects running on the hardware
        serial = self.device.serial
        state = procpid.DeviceSoftwareState(serial)
        process = procpid.ProcessManager(serial)
        if state.get_effect() or process.is_another_instance_is_running():
            process.stop()

        self.dbg.stdout("Previewing effect '{0}' on device '{1}'.".format(self.data["name"], device_name), self.dbg.success, 1)

    def _init_playback_controls(self):
        """
        Connect the signals for playing back sequence effects within the editor.
        For sequence effects only.
        """
        self.playback_paused = True
        self.playback_stop.setVisible(False)
        self.btn_playback_stop.setVisible(False)

        class PlaybackThread(QThread):
            """
            Playing back sequence effects uses the existing editor signals and
            implementation to automatically go through each frame until the
            user wishes to stop.
            """
            @staticmethod
            def run():
                total_frames = self.frame_table.columnCount() - 1
                while not self.playback_paused and self.alive:
                    self.frame_table.selectColumn(self.frame_table.currentColumn() + 1)
                    self.open_frame(True)

                    if self.current_frame == total_frames:
                        if self.data["loop"] == True:
                            time.sleep(1 / self.data["fps"])
                            self.frame_table.selectColumn(0)
                            self.open_frame(True)
                        else:
                            return

                    time.sleep(1 / self.data["fps"])

        def _jump_start():
            self.frame_table.selectColumn(0)

        def _jump_end():
            self.frame_table.selectColumn(self.frame_table.columnCount() - 1)

        def _jump_prev():
            self.frame_table.selectColumn(self.frame_table.currentColumn() - 1)

        def _jump_next():
            self.frame_table.selectColumn(self.frame_table.currentColumn() + 1)

        def _play_pause_ui(paused):
            self.playback_paused = paused
            self.playback_stop.setVisible(not paused)
            self.btn_playback_stop.setVisible(not paused)
            self.playback_play.setVisible(paused)
            self.btn_playback_play.setVisible(paused)
            self._update_disabled_frame_controls()

        def _play():
            _jump_start()
            self.playback_thread.start()
            self.dbg.stdout("Playback started", self.dbg.action, 1)
            _play_pause_ui(False)

        def _stop():
            self.playback_thread.exit()

            if not self.alive:
                return

            self.dbg.stdout("Playback stopped", self.dbg.action, 1)
            _play_pause_ui(True)

        def _toggle_loop(checked):
            self.set_modified(True)
            self.dbg.stdout("Sequence set to {0}".format("loop" if checked else "play once"), self.dbg.debug, 1)
            self.data["loop"] = checked
            self.playback_loop.setChecked(checked)
            self.btn_playback_loop.setChecked(checked)

        def _set_fps(new_fps):
            self.set_modified(True)
            self.dbg.stdout("Sequence frame rate set to: {0} fps".format(new_fps), self.dbg.debug, 1)
            self.data["fps"] = new_fps

        self.playback_jump_start.triggered.connect(_jump_start)
        self.playback_jump_end.triggered.connect(_jump_end)
        self.playback_prev.triggered.connect(_jump_prev)
        self.playback_next.triggered.connect(_jump_next)
        self.playback_play.triggered.connect(_play)
        self.playback_stop.triggered.connect(_stop)
        self.playback_loop.triggered.connect(_toggle_loop)

        self.btn_playback_jump_start.clicked.connect(_jump_start)
        self.btn_playback_jump_end.clicked.connect(_jump_end)
        self.btn_playback_prev.clicked.connect(_jump_prev)
        self.btn_playback_next.clicked.connect(_jump_next)
        self.btn_playback_play.clicked.connect(_play)
        self.btn_playback_stop.clicked.connect(_stop)
        self.btn_playback_loop.clicked.connect(_toggle_loop)
        self.spinner_playback_fps.valueChanged.connect(_set_fps)

        # Set initial values
        self.playback_loop.setChecked(self.data["loop"] == True)
        self.btn_playback_loop.setChecked(self.data["loop"] == True)
        self.spinner_playback_fps.setValue(int(self.data["fps"]))

        # Prepare thread to playback the effect
        self.playback_thread = PlaybackThread()
        self.playback_thread.finished.connect(_stop)

    def closeEvent(self, event=None):
        """
        Gracefully closes the editor, ensuring any unsaved changes are saved
        if desired, and that any thread(s) are stopped.
        """
        # Prompt for unsaved changes
        if self.modified:
            def _do_not_save():
                self.modified = False
                return self.closeEvent()

            def _do_save():
                return self.save_and_close()

            self.widgets.open_dialog(self.widgets.dialog_warning,
                                     self._("Unsaved Changes"),
                                     self._("This effect was modified. Would you like to save these changes?"),
                                     buttons=[QMessageBox.Save, QMessageBox.Discard, QMessageBox.Cancel],
                                     default_button=QMessageBox.Save,
                                     actions={
                                        QMessageBox.Save: _do_save,
                                        QMessageBox.Discard: _do_not_save
                                     })
            if event:
                event.ignore()
            return

        # Stop any threads
        if self.playback_thread:
            self.dbg.stdout("Stopping playback thread...", self.dbg.action, 1)
            self.playback_thread.exit()

        # If live preview was active, restore original state
        if self.live_preview and self.device:
            self.dbg.stdout("Restoring original device state...", self.dbg.action, 1)
            for zone in self.device.zones:
                self.middleman.replay_active_effect(self.device)

        # Window geometry
        self.appdata.main_window._save_window_position(self.window, "editor")

        # Close window and allow another editor to open again
        self.alive = False
        if event:
            event.accept()
        else:
            self.window.deleteLater()

    def set_modified(self, state):
        """
        Sets the modified flag for this editor. The window title is updated
        accordingly to display the current file and its state.
        """
        self.modified = state
        self.window.setWindowTitle("{0}{2} — {1}".format(self.data["name"], self._("Effect Editor"), "*" if state else ""))

    def _get_layer_labels(self):
        """
        Return a dictionary of human readable labels for layered effects.
        """
        return {
            effects.LAYER_STATIC: self._("Static"),
            effects.LAYER_GRADIENT: self._("Gradient"),
            effects.LAYER_PULSING: self._("Pulsing"),
            effects.LAYER_WAVE: self._("Wave"),
            effects.LAYER_SPECTRUM: self._("Spectrum"),
            effects.LAYER_CYCLE: self._("Cycling"),
            effects.LAYER_SCRIPT: self._("Run Scripted Effect")
        }

    def _show_file_error(self):
        """
        Alias to the _show_file_error() function from the Effects tab.
        """
        self.appdata.tab_effects._show_file_error()
        self.closeEvent()

    def save_and_close(self):
        """
        Saves the data in memory to disk, then closes the editor.
        """
        success = self.save_now()

        if success:
            self.closeEvent()

    def save_now(self):
        """
        Saves the data in memory to disk.
        """
        self.dbg.stdout("Saving Effect: " + self.save_path, self.dbg.action, 1)
        success, save_path = self.fileman.save_item(self.data, self.save_path)

        if not success:
            self.widgets.open_dialog(self.widgets.dialog_error,
                                     self._("Save Error"),
                                     self._("Save failed. Please check the permissions and try again."))
            return False

        self.set_modified(False)
        self.save_path = save_path
        self.statusbar.showMessage(self._("Saved to: []").replace("[]", save_path))
        return True

    def reload_from_file(self):
        """
        Prompts the user if they're happy to dispose of the data in memory in
        favour of the data stored on disk.
        """
        def _do_reload():
            """
            Hot reload the editor from the beginning via the effects class.
            """
            self.dbg.stdout("Reloading effect from disk...", self.dbg.action, 1)
            self.modified = False
            self.alive = False
            self.appdata.tab_effects.open_file(self.save_path)
            self.appdata.tab_effects.edit_file()
            self.closeEvent()

        self.widgets.open_dialog(self.widgets.dialog_warning,
                                 self._("Revert"),
                                 self._("Reload the effect from disk? Any unsaved data will be lost!"),
                                 buttons=[QMessageBox.Yes, QMessageBox.No],
                                 default_button=QMessageBox.Yes,
                                 actions={QMessageBox.Yes: _do_reload})

    # ----------------------------
    # Common
    # ----------------------------
    def _redraw_editor(self):
        """
        Alias to redraw the effect on the editor, based on the effect type.
        """
        aliases = {
            effects.TYPE_LAYERED: self.open_layer,
            effects.TYPE_SEQUENCE: self.open_frame
        }
        aliases[self.effect_type]()

    def _shift_all_positions(self, relative_x, relative_y):
        """
        Offsets the positions of LEDs by the specified relative X or Y value.

        This function only supports the movement of 1 column/row of 1 direction
        at a time.
        """
        self.set_modified(True)

        # Load data into memory
        if self.layered_effect:
            print("fixme:_shift_all_positions TYPE_LAYERED")
            raise NotImplementedError

        elif self.sequence_effect:
            frame = self.data["frames"][self.current_frame]
            new_frame = {}

        # Determine operations
        move_left = relative_x < 0
        move_right = relative_x > 0
        move_down = relative_y > 0
        move_up = relative_y < 0

        # Process changes
        max_x = self.data["map_cols"] - 1
        max_y = self.data["map_rows"] - 1

        if move_left or move_right:
            for x in frame.keys():
                x = int(x)

                # Shift X-axis
                if move_left:
                    new_frame[str(x - 1)] = frame[str(x)]
                elif move_right:
                    new_frame[str(x + 1)] = frame[str(x)]

                # Wrap X-axis
                new_x_keys = new_frame.keys()

                # -- First item goes last
                if str(-1) in new_x_keys:
                    new_frame[str(max_x)] = frame[str(0)]
                    del(new_frame[str(-1)])

                # -- Last item goes first
                if str(max_x + 1) in new_x_keys:
                    new_frame[str(0)] = frame[str(max_x)]
                    del(new_frame[str(max_x + 1)])

        elif move_up or move_down:
            for x in frame.keys():
                x = int(x)
                y_keys = list(frame[str(x)].keys())
                new_frame[str(x)] = {}

                for y in y_keys:
                    y = int(y)

                    # Shift Y-axis
                    if move_up:
                        new_frame[str(x)][str(y - 1)] = frame[str(x)][str(y)]
                    elif move_down:
                        new_frame[str(x)][str(y + 1)] = frame[str(x)][str(y)]

                    # Wrap Y-axis
                    new_y_keys = new_frame[str(x)].keys()

                    # -- First item goes last
                    if str(-1) in new_y_keys:
                        new_frame[str(x)][str(max_y)] = frame[str(x)][str(0)]
                        del(new_frame[str(x)][str(-1)])

                    # -- Last item goes first
                    if str(max_y + 1) in new_y_keys:
                        new_frame[str(x)][str(0)] = frame[str(x)][str(max_y)]
                        del(new_frame[str(x)][str(max_y + 1)])

        # Save new data and refresh UI
        if self.layered_effect:
            print("fixme:_shift_all_positions TYPE_LAYERED")
            raise NotImplementedError

        elif self.sequence_effect:
            self.data["frames"][self.current_frame] = new_frame
            self.open_frame()

    def shift_all_left(self):
        """
        Move all the LEDs by -1 on the X-axis.
        """
        self._shift_all_positions(-1, 0)

    def shift_all_right(self):
        """
        Move all the LEDs by +1 on the X-axis.
        """
        self._shift_all_positions(1, 0)

    def shift_all_up(self):
        """
        Move all the LEDs by -1 on the Y-axis.
        """
        self._shift_all_positions(0, -1)

    def shift_all_down(self):
        """
        Move all the LEDs by +1 on the Y-axis.
        """
        self._shift_all_positions(0, 1)

    def edit_metadata(self):
        """
        Opens a modal metadata window and reloads the effect upon accepting the
        changes.
        """
        def _metadata_changed(newdata):
            if newdata == self.data:
                self.dbg.stdout("Metadata unchanged.", self.dbg.success, 1)
                return

            self.dbg.stdout("Metadata changed, refreshing...", self.dbg.action, 1)
            self.action_view_device_graphic.setDisabled(True if not newdata["map_graphic"] else False)
            self.set_modified(True)
            self.data = newdata
            self.view_device_graphic()
            self.statusbar.clearMessage()

        def _metadata_closed(result):
            self.window.setEnabled(True)

        self.window.setEnabled(False)
        self.metadata_editor = controller_effects.EffectMetadataEditor(self.appdata, self.data, _metadata_changed)
        self.metadata_editor.dialog.finished.connect(_metadata_closed)

    def edit_triggers(self):
        """
        Opens a window to quickly create, modify or delete a trigger that
        uses this effect.
        """
        pass

    def _live_preview_failed(self, e):
        """
        Inform the user the live preview failed.

        Params:
            e       Raw exception
        """
        self.dbg.stdout("Failed to preview on device! Exception: ", self.dbg.error)
        self.dbg.stdout(common.get_exception_as_string(e) + "\n", self.dbg.error)

        # Disable live preview for session
        self.device = None
        self.matrix = None
        self.widgets.open_dialog(self.widgets.dialog_error,
                                 self._("Backend Error"),
                                 self._("An error occurred while sending the frame to the hardware."),
                                 info_text=self._("Live preview has been disabled for this session."),
                                 details=common.get_exception_as_string(e))

    # ----------------------------
    # Tools
    # ----------------------------
    def _select_mode_common(self):
        self.tool_draw.setChecked(False)
        self.tool_eraser.setChecked(False)
        self.tool_picker.setChecked(False)

    def select_mode_draw(self):
        """
        Select tool 1 for the visual editor.

            Layered     Assign LEDs/keys to the current layer
            Sequence    Draw colours on LEDs/keys
        """
        self.current_tool = VISUAL_MODE_ADD
        self.device_renderer.set_mode(VISUAL_MODE_ADD)
        self._select_mode_common()
        self.tool_draw.setChecked(True)

    def select_mode_eraser(self):
        """
        Select tool 2 for the visual editor.

            Layered     Unassign LEDs/keys from the current layer
            Sequence    Clear colour on LEDs/keys
        """
        self.current_tool = VISUAL_MODE_DEL
        self.device_renderer.set_mode(VISUAL_MODE_DEL)
        self._select_mode_common()
        self.tool_eraser.setChecked(True)

    def select_mode_picker(self):
        """
        Select tool 3 for the visual editor.

            Layered     n/a - Unused
            Sequence    Grab the colour on an LED/key
        """
        self.current_tool = VISUAL_MODE_PICK
        self.device_renderer.set_mode(VISUAL_MODE_PICK)
        self._select_mode_common()
        self.tool_picker.setChecked(True)

    # ----------------------------
    # Visual Editor (WebView)
    # ----------------------------
    def _update_zoom_controls(self):
        """
        Enable/disable the zoom controls accordingly, to avoid going under or
        over the min/max zoom levels
        """
        self.webview.setZoomFactor(self.webview_zoom_level)
        min_zoom = 0.25
        max_zoom = 3

        self.action_zoom_in.setEnabled(True)
        self.action_zoom_out.setEnabled(True)

        if self.webview_zoom_level <= min_zoom:
            self.action_zoom_out.setEnabled(False)

        if self.webview_zoom_level >= max_zoom:
            self.action_zoom_in.setEnabled(False)

    def zoom_in(self):
        """
        Increase the zoom of the visual editor viewport.
        """
        self.webview_zoom_level += 0.25
        self._update_zoom_controls()

    def zoom_out(self):
        """
        Decrease the zoom of the visual editor viewport.
        """
        self.webview_zoom_level -= 0.25
        self._update_zoom_controls()

    def zoom_reset(self):
        """
        Reset the zoom of the visual editor viewport.
        """
        self.webview_zoom_level = 1.0
        self._update_zoom_controls()

    def _view_device_reload_ready(self):
        """
        Shared private function for when the user changes the graphic while
        the editor is open. This ensures the state is loaded properly.
        """
        self.init_editor()
        self.device_renderer.set_colour(self.current_colour)
        self.device_renderer.set_mode(self.current_tool)
        self.statusbar.showMessage(self._("Temporarily changed the graphic. To make permanent, edit the metadata."), 5000)

    def view_device_graphic(self):
        """
        Temporarily switch the visual editor graphic to the hardware graphic,
        if available.
        """
        self.statusbar.showMessage(self._("Loading..."), 5000)
        hide_key_labels = not self.action_view_key_labels.isChecked()
        self.device_renderer = DeviceRenderer(self.appdata, self, self.webview, self._view_device_reload_ready, self.data["map_graphic"], self.data["map_rows"], self.data["map_cols"], hide_key_labels)

        if self.data["map_device_icon"] in ["keyboard", "keypad"]:
            self.action_view_key_labels.setEnabled(True)

    def view_device_grid(self):
        """
        Temporarily switch the visual editor graphic to a grid, which allows
        drawing on any device.
        """
        self.statusbar.showMessage(self._("Loading..."), 5000)
        hide_key_labels = not self.action_view_key_labels.isChecked()
        self.device_renderer = DeviceRenderer(self.appdata, self, self.webview, self._view_device_reload_ready, "", self.data["map_rows"], self.data["map_cols"], hide_key_labels)

        self.action_view_key_labels.setEnabled(False)

    def view_key_labels(self):
        """
        Temporarily show/hide the key labels (for keyboards and keypads only)
        """
        if self.action_view_device_graphic.isChecked():
            self.view_device_graphic()
        else:
            self.view_device_grid()
        self.statusbar.showMessage(self._("Temporarily changed the graphic. To make permanent, edit the application's preferences."), 5000)

    # ----------------------------
    # Specific to Layered Effects
    # ----------------------------
    def load_layers(self):
        """
        Clears and populates layers on the dock controls.
        """
        pass

    def open_layer(self):
        """
        Draw the selected layer for editing, and update the properties dock.

        This will redraw the colours in the visual editor and physical device
        (if preview is enabled) based on the currently selected layer.
        """
        pass

    def new_layer(self):
        """
        Create a new layer using default values and loads it.
        """
        pass

    def delete_layer(self):
        """
        Prompts for confirmation before deleting a layer.
        """
        pass

    def duplicate_layer(self):
        """
        Creates a new layer inheriting the data for the currently selected one
        and loads it.
        """
        pass

    def raise_layer(self):
        """
        Moves the layer up by one and refreshes the visual editor.
        """
        pass

    def lower_layer(self):
        """
        Moves the layer down by one and refreshes the visual editor.
        """
        pass

    def assign_key_to_layer(self, x, y):
        """
        User clicks on a key/LED in the visual editor. Add this to the layer
        and update the device preview.
        """
        pass

    def unassign_key_to_layer(self, x, y):
        """
        User clicks on a key/LED in the visual editor. Remove this from the
        layer and update the device preview.
        """
        pass

    # ----------------------------
    # Specific to Sequence Effects
    # ----------------------------
    def load_frames(self):
        """
        Clears and populates frames on the dock controls.
        """
        table = self.frame_table
        frame_count = len(self.data["frames"])

        table.setColumnCount(frame_count)
        for frame in range(0, frame_count):
            table.setColumnWidth(frame, 64)

        # Make sure there is one frame
        if frame_count == 0:
            table.insertColumn(0)
            self.data["frames"] = [{}]

        # Select first frame if deselected
        if len(table.selectedIndexes()) == 0:
            table.setSelection(QRect(0, 0, 1, 1), QItemSelectionModel.Select)

    def open_frame(self, playback_mode=False):
        """
        Open the selected frame for editing.

        This will redraw the colours in the visual editor and physical device
        (if preview is enabled) based on the currently selected frame.

        When 'playback_mode' is on, UI controls will not be updated.
        """
        fx = self.matrix
        self.device_renderer.clear()

        if fx:
            fx.clear()

        if not self.frame_table.selectedIndexes():
            self.frame_table.selectColumn(self.current_frame)
            return

        index = self.frame_table.selectedIndexes()[0].column()
        self.current_frame = index
        total_X = self.data["map_cols"]
        total_Y = self.data["map_rows"]

        # Make sure the frame dictionary keys have parent (x-axis) keys
        for x in range(0, total_X):
            try:
                self.data["frames"][index][str(x)]
            except KeyError:
                # New frame to be written
                self.data["frames"][index][str(x)] = {}
            except IndexError:
                # The last index was just deleted
                self.frame_table.selectColumn(self.current_frame)
                return

        self.status_frame.setText(self._("Frame [1] of [10]").replace("[1]", str(self.current_frame + 1)).replace("[10]", str(self.frame_table.columnCount())))

        # Draw frame on editor (and optionally physical hardware)
        frame = self.data["frames"][index]
        for x in frame.keys():
            for y in frame[x].keys():
                try:
                    hex_value = frame[str(x)][str(y)]
                    self.device_renderer.set_pos(x, y, hex_value)
                except KeyError:
                    # Expected, no data stored for this position
                    continue

                if fx:
                    try:
                        rgb = fx.hex_to_rgb(hex_value)
                        fx.set(int(x), int(y), rgb[0], rgb[1], rgb[2])
                    except Exception as e:
                        self._live_preview_failed(e)

        if fx:
            fx.draw()

        if not playback_mode:
            self._update_disabled_frame_controls()

    def _update_disabled_frame_controls(self):
        """
        Reusable code to update the state of controls.

        For example, disable buttons that are impossible (e.g. move to right
        when already at the end) as well as preventing editing while
        the animation is playing.
        """
        # -- Restricted during playback
        for widget in [
                self.webview, self.dock_colours,
                self.tool_draw, self.tool_eraser, self.tool_picker,
                self.action_edit_metadata, self.action_edit_triggers,

                self.frame_table,
                self.action_new_frame, self.btn_frame_new,
                self.action_clone_frame, self.btn_frame_clone,

                self.menu_shift, self.menu_edit, self.menu_tools, self.menu_colours,
                self.action_shift_left, self.action_shift_right,
                self.action_shift_up, self.action_shift_down
            ]:
            widget.setEnabled(self.playback_paused)

        # -- Prevent deleting the only frame
        total_frames = len(self.data["frames"])
        self.action_delete_frame.setEnabled(total_frames > 1)
        self.btn_frame_delete.setEnabled(total_frames > 1)

        # -- Preventing shift/step frames at start/end
        self.btn_frame_move_left.setEnabled(self.current_frame >= 1)
        self.action_frame_left.setEnabled(self.current_frame >= 1)
        self.playback_prev.setEnabled(self.current_frame >= 1)
        self.btn_playback_prev.setEnabled(self.current_frame >= 1)
        self.playback_jump_start.setEnabled(self.current_frame >= 1)
        self.btn_playback_jump_start.setEnabled(self.current_frame >= 1)

        self.btn_frame_move_right.setEnabled(self.current_frame < total_frames - 1)
        self.action_frame_right.setEnabled(self.current_frame < total_frames - 1)
        self.playback_next.setEnabled(self.current_frame < total_frames - 1)
        self.btn_playback_next.setEnabled(self.current_frame < total_frames - 1)
        self.playback_jump_end.setEnabled(self.current_frame < total_frames - 1)
        self.btn_playback_jump_end.setEnabled(self.current_frame < total_frames - 1)

        # -- Restricted during playback but frame specific
        if not self.playback_paused:
            for widget in [
                self.action_delete_frame, self.btn_frame_delete,
                self.action_frame_left, self.btn_frame_move_left,
                self.action_frame_right, self.btn_frame_move_right,
                self.playback_prev, self.btn_playback_prev,
                self.playback_next, self.btn_playback_next,
                self.playback_jump_start, self.btn_playback_jump_start,
                self.playback_jump_end, self.btn_playback_jump_end
            ]:
                widget.setEnabled(False)

    def load_colours(self):
        """
        Populate the colour dock and functions associated for quickly setting
        the brush (draw) colour.

        Intended for use with sequence effects, the "pixel art" style of editing.
        In addition, saved colours will have lighter/darker colours generated.
        """
        colours = pref.load_file(self.paths.colours)

        # Current colour (picker) and label
        def _set_custom_colour(new_hex, data):
            self._set_current_colour(new_hex)

            # Regenerate the colour list if the list has changed.
            if colours != pref.load_file(self.paths.colours):
                self.dbg.stdout("Colour list changed, reloading dock...", self.dbg.action, 1)
                self.load_colours()

        # Populate saved colours (and variants)
        def _mk_button(parent_widget, name, hex_value):
            button = QPushButton()
            button.setIconSize(QSize(32, 32))
            button.resize(QSize(32, 32))
            button.setFixedSize(32, 32)
            button.setFlat(True)
            button.setToolTip("{0} ({1})".format(name, hex_value))
            button.setStyleSheet("QPushButton { border: none; background: none; padding: 0; margin: 0; }")
            button.setIcon(QIcon(common.generate_colour_bitmap(self.dbg, hex_value, 32)))
            button.clicked.connect(lambda a: self._set_current_colour(hex_value))
            parent_widget.layout().addWidget(button)

        container = QWidget()
        container.setLayout(QFlowLayout())
        container.layout().setSpacing(2)
        container.layout().setContentsMargins(0, 0, 0, 0)

        edit_btn = QPushButton()
        if not self.appdata.system_qt_theme:
            edit_btn.setIcon(self.widgets.get_icon_qt("general", "edit"))
        edit_btn.setText(self._("Edit"))
        edit_btn.setStatusTip(self._("Modify or reorder your saved colours"))
        edit_btn.clicked.connect(self.edit_saved_colours)

        for colour in colours:
            shades = [-45, -30, -15, 0, 15, 30]
            if not self.preferences["editor"]["show_saved_colour_shades"]:
                shades = [0]
            for percent in shades:
                percent = float(percent / 100)
                new_colour = self.fx.lightness_hex(colour["hex"], percent)
                label = "{0} {1}%".format(colour["name"], int(percent * 100))
                if percent == 0:
                    label = colour["name"]
                _mk_button(container, label, new_colour)

        shared.clear_layout(self.saved_colours.layout())
        self.saved_colours.layout().addWidget(container)
        self.saved_colours.layout().addWidget(edit_btn)
        self.dock_colours.setMinimumWidth(240)

    def new_frame(self):
        """
        Creates a blank frame after the current frame, and loads it.
        """
        self.set_modified(True)

        current_index = self.current_frame
        new_index = self.current_frame + 1
        self.dbg.stdout("New frame at position {0}".format(new_index), self.dbg.debug, 1)

        self.data["frames"].insert(new_index, {})

        self.frame_table.insertColumn(new_index)
        self.frame_table.setColumnWidth(new_index, 64)
        self.frame_table.selectColumn(new_index)

    def delete_frame(self):
        """
        Prompts for confirmation before deleting a frame.
        """
        def _do_delete_frame():
            self.set_modified(True)

            old_index = self.current_frame
            new_index = self.current_frame - 1

            if new_index < 0:
                new_index = 0

            del(self.data["frames"][old_index])
            self.frame_table.removeColumn(old_index)
            self.frame_table.selectColumn(new_index)

        if self.suppress_delete_prompt:
            return _do_delete_frame()

        self.widgets.open_dialog(self.widgets.dialog_warning,
                                 self._("Delete Frame"),
                                 self._("Permanently delete this frame? There is no undo."),
                                 info_text=self._("Tip: This confirmation can be suppressed via the Editor tab under Preferences."),
                                 buttons=[QMessageBox.Yes, QMessageBox.No],
                                 default_button=QMessageBox.Yes,
                                 actions={QMessageBox.Yes: _do_delete_frame})

    def clone_frame(self):
        """
        Creates a new frame inheriting the data for the currently selected one
        and loads it.
        """
        self.set_modified(True)

        current_index = self.current_frame
        new_index = self.current_frame + 1
        self.dbg.stdout("Cloned frame at position {0}".format(new_index), self.dbg.debug, 1)

        newdata = copy.deepcopy(self.data["frames"][current_index])
        self.data["frames"].insert(new_index, newdata)

        self.frame_table.insertColumn(new_index)
        self.frame_table.setColumnWidth(new_index, 64)
        self.frame_table.selectColumn(new_index)

    def _swap_frame_data(self, relative_pos):
        """
        Relocate frame data with another data position and reload the frame.
        """
        self.set_modified(True)

        current_index = self.current_frame
        new_index = self.current_frame + relative_pos

        data_src = copy.deepcopy(self.data["frames"][current_index])
        data_dst = copy.deepcopy(self.data["frames"][new_index])

        self.data["frames"][current_index] = data_dst
        self.data["frames"][new_index] = data_src

        self.dbg.stdout("Swapped frame {0} with frame {1}".format(current_index, new_index), self.dbg.debug, 1)
        self.frame_table.selectColumn(new_index)

    def shift_frame_left(self):
        """
        Move the frame to the left by one and refresh the visual editor.
        """
        self._swap_frame_data(-1)

    def shift_frame_right(self):
        """
        Move the frame to the right by one and refresh the visual editor.
        """
        self._swap_frame_data(1)

    def _set_current_colour(self, hex_value, no_textbox_update=False):
        """
        Change the colour to draw with. If the draw tool isn't selected,
        this will automatically change.
        """
        self.dbg.stdout("Brush colour changed: " + hex_value, self.dbg.debug, 1)
        self.current_colour = hex_value
        self.device_renderer.set_colour(hex_value)
        self.select_mode_draw()

        # Allows the user to keep typing when directly typing hex value
        if not no_textbox_update:
            self.current_colour_textbox.setText(hex_value)

        self.current_colour_block.setStyleSheet("QWidget {{ background-color: {0} }}".format(hex_value))
        self._refresh_colour_tweak_controls()

    def open_colour_picker(self):
        """
        Opens the system (Qt) colour picker for choosing a new colour on the fly.
        """
        dialog = QColorDialog()

        # TODO: Nice to have: Use Polychromatic styling for the dialog
        # QColorDialog() is private, so this may be tricky.
        # This does not work!
        shared.load_qt_theme(self.appdata, dialog)

        output = dialog.getColor(title=self._("Choose Brush Colour"), initial=QColor(self.current_colour))

        if not output.isValid():
            return
        self._set_current_colour(output.name())

    def edit_saved_colours(self):
        """
        Opens Polychromatic's colour picker for managing the saved colour list.
        """
        def _update_saved_colours(new_hex, params=[]):
            self._set_current_colour(new_hex)
            self.load_colours()

        picker = self.widgets.create_colour_control(self.current_colour, _update_saved_colours, None, self._("Saved Colours"))
        picker.findChild(QPushButton).click()

    def draw_LED_to_frame(self, x, y):
        """
        User draws a new position on the current frame.
        """
        def _release_click():
            """After an error dialog, the button ends up "stuck" in the webview."""
            self.webview.page().runJavaScript("mouseDown = 0;")

        try:
            self.data["frames"][self.current_frame][str(x)][str(y)] = self.current_colour
        except KeyError:
            _release_click()
            self.dbg.stdout("Can't set LED ({0}, {1}) as out of bounds for device (or save data)".format(str(x), str(y)), self.dbg.error)
            self.widgets.open_dialog(self.widgets.dialog_error,
                                     self._("Out of Bounds"),
                                     self._("There isn't a position for (X,Y) for this device. The graphic contains incorrect metadata or is incompatible for use with this device.").replace("(X,Y)", "({0},{1})".format(str(x), str(y))))
            return

        if self.matrix:
            try:
                rgb = common.hex_to_rgb(self.current_colour)
                self.matrix.set(int(x), int(y), rgb[0], rgb[1], rgb[2])
                self.matrix.draw()
            except Exception as e:
                _release_click()
                self._live_preview_failed(e)

        self.set_modified(True)
        self.dbg.stdout("Set LED ({0},{1}) to {2}".format(x, y, self.current_colour), self.dbg.debug, 1)

    def erase_LED_from_frame(self, x, y):
        """
        User erases a position on the current frame.
        """
        try:
            del(self.data["frames"][self.current_frame][str(x)][str(y)])
        except KeyError:
            self.statusbar.showMessage(self._("This LED is empty - nothing to erase here!"), 5000)
            return

        if self.matrix:
            try:
                self.matrix.set(int(x), int(y), 0, 0, 0)
                self.matrix.draw()
            except Exception as e:
                self._live_preview_failed(e)

        self.set_modified(True)
        self.dbg.stdout("Erase LED ({0},{1})".format(x, y), self.dbg.debug, 1)

    def pick_LED_colour_from_frame(self, x, y):
        """
        User picks a colour from the current frame.
        """
        try:
            hex_value = self.data["frames"][self.current_frame][str(x)][str(y)]
            self._set_current_colour(hex_value)
        except KeyError:
            self.statusbar.showMessage(self._("This LED is empty - no colour to pick here!"), 5000)
            return

        self.dbg.stdout("Pick LED ({0},{1}). Current colour changed to {2}".format(x, y, self.current_colour), self.dbg.debug, 1)

    def set_colour_by_textbox(self, new_hex):
        """
        Emitted when the hex value is changed. The current colour will only be
        updated when the full hex is specified. Qt's input mask already validated this.
        """
        if len(new_hex) < 7:
            return
        self._set_current_colour(new_hex, no_textbox_update=True)

    def _refresh_colour_tweak_controls(self):
        """
        Enables/disables controls for HSL adjustments when the next/previous value
        has been exhausted. For example, pure white (#fff) cannot get any lighter.
        """
        for widget in [self.btn_hue_increase, self.btn_hue_decrease,
                       self.btn_saturation_increase, self.btn_saturation_decrease,
                       self.btn_lightness_decrease, self.btn_lightness_increase,
                       self.action_colours_hue_increase, self.action_colours_hue_decrease,
                       self.action_colours_saturation_increase, self.action_colours_saturation_decrease,
                       self.action_colours_lightness_increase, self.action_colours_lightness_decrease]:
            widget.setEnabled(True)

        if self.fx.hue_hex(self.current_colour, 0.05) == self.current_colour:
            self.btn_hue_increase.setEnabled(False)
            self.action_colours_hue_increase.setEnabled(False)

        if self.fx.hue_hex(self.current_colour, -0.05) == self.current_colour:
            self.btn_hue_decrease.setEnabled(False)
            self.action_colours_hue_decrease.setEnabled(False)

        if self.fx.saturate_hex(self.current_colour, 0.05) == self.current_colour:
            self.btn_saturation_increase.setEnabled(False)
            self.action_colours_saturation_increase.setEnabled(False)

        if self.fx.saturate_hex(self.current_colour, -0.05) == self.current_colour:
            self.btn_saturation_decrease.setEnabled(False)
            self.action_colours_saturation_decrease.setEnabled(False)

        if self.fx.lightness_hex(self.current_colour, 0.05) == self.current_colour:
            self.btn_lightness_increase.setEnabled(False)
            self.action_colours_lightness_increase.setEnabled(False)

        if self.fx.lightness_hex(self.current_colour, -0.05) == self.current_colour:
            self.btn_lightness_decrease.setEnabled(False)
            self.action_colours_lightness_decrease.setEnabled(False)

    def set_colour_increase_hue(self):
        """
        Increases the Hue (HSL) of the current colour.
        """
        self._set_current_colour(self.fx.hue_hex(self.current_colour, 0.05))
        self._refresh_colour_tweak_controls()

    def set_colour_decrease_hue(self):
        """
        Decreases the Hue (HSL) of the current colour.
        """
        self._set_current_colour(self.fx.hue_hex(self.current_colour, -0.05))
        self._refresh_colour_tweak_controls()

    def set_colour_increase_saturation(self):
        """
        Increases the Saturation (HSL) of the current colour.
        """
        self._set_current_colour(self.fx.saturate_hex(self.current_colour, 0.05))
        self._refresh_colour_tweak_controls()

    def set_colour_decrease_saturation(self):
        """
        Decreases the Saturation (HSL) of the current colour.
        """
        self._set_current_colour(self.fx.saturate_hex(self.current_colour, -0.05))
        self._refresh_colour_tweak_controls()

    def set_colour_increase_lightness(self):
        """
        Increases the Value (HSL) of the current colour.
        """
        self._set_current_colour(self.fx.lightness_hex(self.current_colour, 0.05))
        self._refresh_colour_tweak_controls()

    def set_colour_decrease_lightness(self):
        """
        Decreases the Value (HSL) of the current colour.
        """
        self._set_current_colour(self.fx.lightness_hex(self.current_colour, -0.05))
        self._refresh_colour_tweak_controls()


class DeviceRenderer(shared.TabData):
    """
    Responsible for the input/output of the device graphic or grid layout.
    """
    def __init__(self, appdata, editor, webview, ready_fn, map_graphic, map_rows, map_cols, hide_labels=True):
        """
        Params:
            appdata         ApplicationData() object
            editor          VisualEffectEditor() object
            webview         QWebEngineView object
            ready_fn        Run this function as soon as the editor has loaded
            map_graphic     Filename of graphic to use. Blank string indicates to use grid.
            map_rows        Number of rows for device
            map_cols        Number of cols for device
            show_labels     Boolean of whether labels on keyboards/keypads should be visible
        """
        self.appdata = appdata
        super().__init__(appdata)
        self.editor = editor
        self.webview = webview
        self.loaded = False
        self.ready_fn = ready_fn
        self.use_native_cursor = appdata.preferences["editor"]["system_cursors"]
        self.hide_key_labels = hide_labels

        self.device_map = effects.DeviceMapGraphics(appdata)
        self.graphic_filename = map_graphic
        self.graphic_name = self.device_map.get_graphic_name_from_filename(map_graphic)
        self.rows = map_rows
        self.cols = map_cols
        self.mode = 1

        # Wait until the webview has fully loaded, then load the graphic.
        self.webview.loadFinished.connect(self.ready)
        url = QUrl("file://" + os.path.join(self.paths.data_dir, "qt", "editor.html"))

        # Reload if web view previously initialised
        if self.webview.page().url() == url:
            self.webview.reload()
        else:
            # First time opening this editor session
            self.webview.load(url)

        # Connect signals
        self.webview.titleChanged.connect(self._cb_title_changed)

    def ready(self, ok=True):
        """
        As soon as the WebView has fully loaded, load the graphic ready for
        editing, as well as the ready_fn.
        """
        self.loaded = True
        if not ok:
            self.widgets.open_dialog(self.widgets.dialog_error,
                                     self._("Controller Error"),
                                     self._("Failed to load the Qt Web Engine. Unfortunately, editing won't be possible."),
                                     info_text=self._("Make sure the application (and its dependencies) are installed and configured correctly."))
            self.editor.window.close()

        # Load the SVG into the viewport
        svg = self._get_graphic_svg().replace("'", "&quot;")
        self.webview.page().runJavaScript("loadSVG('{0}'); clearLED()".format(svg))

        # Set mode
        self.set_mode(VISUAL_MODE_ADD)

        # Remove scroll bar theming if user prefers native (although it'll look like Chromium instead)
        # Also sets the background colour to match native controls.
        if self.appdata.system_qt_theme:
            self.webview.page().runJavaScript("document.querySelector('#polychromatic-qt-theme').remove()")

            bg_colour = self.main_window.palette().alternateBase().color().name()
            self.webview.page().runJavaScript("document.body.style.background = '{0}'".format(bg_colour))

        # Run ready function
        if self.ready_fn:
            self.ready_fn()

    def _get_graphic_svg(self):
        """
        Returns the SVG of the device's matrix represented as a 'pretty' graphic.
        """
        if not self.graphic_filename:
            return self._generate_grid_svg()

        # Update UI elements
        self.editor.action_view_device_graphic.setChecked(True)
        self.editor.action_view_grid.setChecked(False)

        svg = self.device_map.get_svg_graphic(self.graphic_filename)

        if not svg:
            self.dbg.stdout("Graphic does not exist: '{0}'. Using grid as fallback!".format(self.graphic_filename), self.dbg.error)
            self.widgets.open_dialog(self.widgets.dialog_warning,
                                     self._("Missing File"),
                                     self._("This effect is mapped to a graphic named '[]', but it's not installed.".replace("[]", self.graphic_filename)),
                                     info_text=self._("The grid will be used instead. A different graphic can be chosen by editing the metadata."))
            return self._generate_grid_svg()

        if self.hide_key_labels:
            svg = self._hide_key_labels(svg)

        return svg

    def _generate_grid_svg(self):
        """
        Returns an SVG of the device's matrix grid.
        """
        self.editor.action_view_device_graphic.setChecked(False)
        self.editor.action_view_grid.setChecked(True)

        return self.device_map.get_svg_grid(self.cols, self.rows)

    def _cb_title_changed(self, title):
        """
        JavaScript sends string messages to us (Python) by changing the title.
        This will run when the title changes.
        """
        pieces = title.split(";")
        if pieces[0] == "click":
            x = int(pieces[1])
            y = int(pieces[2])
            callbacks = {
                effects.TYPE_SEQUENCE: {
                    VISUAL_MODE_ADD: self.editor.draw_LED_to_frame,
                    VISUAL_MODE_DEL: self.editor.erase_LED_from_frame,
                    VISUAL_MODE_PICK: self.editor.pick_LED_colour_from_frame
                },
                effects.TYPE_LAYERED: {
                    VISUAL_MODE_ADD: self.editor.assign_key_to_layer,
                    VISUAL_MODE_DEL: self.editor.unassign_key_to_layer
                }
            }
            callbacks[self.editor.effect_type][self.editor.current_tool](x, y)

    def set_mode(self, mode):
        """
        Update the cursor/behaviour inside the visual editor, but only when
        the editor is ready.
        """
        if not self.loaded:
            return

        self.mode = mode
        self.webview.page().runJavaScript("setMode({0}, {1})".format(mode, str(self.use_native_cursor).lower()))

    def set_colour(self, hex_value):
        """
        Update the visual editor's current colour in memory.
        """
        self.webview.page().runJavaScript("currentColour = '{0}';".format(hex_value))

        # Reset 'title' so the same command can be sent again (i.e. draw on same position)
        self.webview.page().runJavaScript("document.title = '';")

    def clear(self):
        """
        Clear all the keys.
        """
        if self.loaded:
            self.webview.page().runJavaScript("clearLED()")

    def set_pos(self, x, y, hex_value):
        """
        Set the colour of a specific key.
        """
        self.webview.page().runJavaScript("mode = 1; setLED({0}, {1}, false, '{2}'); mode = {3};".format(x, y, hex_value, self.mode))

    def _hide_key_labels(self, svg):
        """
        Returns an SVG with the "key labels" hidden or removed.

        This is intended for keyboards/keypads which may not have a localized layout
        for the user, but share the same physical layout as an existing one. Users
        may choose to hide labels as a preference too.
        """
        output = []
        lines = svg.splitlines()

        for line in lines:
            if line.find("class=\"label\"") != -1:
                line = line.replace("class=\"label\"", "opacity=\"0\"")
            output.append(line)
        return "\n".join(output)
